' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_SpinVm_004.spin ( Working version - Uses LMM overlays when possible )                    *
' *                                                                                                     *
' *******************************************************************************************************

CON
{

  Bugs fixed from AiChip_SpinVm_003

    Error in ABORT/ABOVAL caused all aborts to trap back to beyond main program causing Cog to go into
    sleep mode. Variable Base Pointer was also left incorrectly aligned on return from RETURN/RETVAL.    
        
  Bugs fixed from AiChip_SpinVm_002

    longVar := ++byteVar : No longer sets longVar to $100 when byteVar is $FF before increment, and
    fixed corresponding --byteVar, and for word vars.

    var := lo #> expr <# hi : Min and Max now used signed operators to do their job properly.

    a := Lockset(), a := LockClr() : Were not returning a result
    
  Changes from AiChip_SpinVm_002

    Bugs detailed above fixed
    Added serial port monitoring
    Abort trapping added
    LookUp and LookDown implemented
    Object handling added
    Stack frame contents slightly changed
    Added 'opcBit' register for Operator and Using so 'opc' register can be used in the called overlay
    to distinguish which opcode got the execution there. This facilitiates merging opcode / operator
    handlers.
    Merged MPY and MPY_BIG
    Merged DIV and MOD
    Merged RPTINCJ and RPTADDJ
    Merged PRE/POST INC/DEC, moved from overlay to shrink the table
    Merged POSTCLR and POSTSET
    Merged WAITPNE and WAITPEQ
    Merged PRESEXB and PRESEXW
    Merged LSETFUN, LSETSUB, LCLRFUN and LCLRSUB
    Relocated Overlay for optimised instruction dispatch in Opcode
    Added LMM-style interpretation of overlay code
    Favoured opcodes and operators moved into main cog ( not overlayed )
    Re-ordered and shrunk Using_table
    Pre-loaded initialisation overlay in overlay area
    Optimised Opr_DoIt and calls to Opr_DoIt
    Optimised 32-bit math in MPY ( 64-bit negation not tested )
    Reduced end of cog padding to a minimum ( $1F0/4 ) so cog address of overlay > $1EF
    Improved disassembler reporting
    Various cosmetic changes
    
  Known or suspected differences with official Parallax Spin Interpreter ...

    Square Root makes negatives absolute first, Parallax returns 'some number'
    Random isn't random, numbers not same as Parallax
    Variables / Memory not zeroed when VM starts
    Stack not used in quite the same way
    Clock not set from details in info block
    Checksum in info block ignored
    PAR is probably different when a Spin program is running

  Not implemented / not working ...

    CogInit, CogNew
    
  Working ( as far as testing so far shows )...

    variables, byte, word, long
    arrays and arrays indexing, var[n], var[n][i], long[n], word[n][i]
    @variable
    Operators, a := a + b
    Assignment operators, a += b
    Pre/Post operation, ++a, a~, ?a
    Unary operators, Absolute, Square Root
    Boolean and comparative operators ( and, => )
    Subroutine and function calls, including nested functions, a := F1(F2(a))
    Subroutine and function parameters
    Objects and object function calls
    local ( stack ) variables    
    Return / Return <expr> / result := <expr>
    Abort trapping / Abort / Abort <expr>
    SPR register bits, DIRA[0], DIRA[7..0] etc, a := ( DIRA[7..4] |= $xx )
    StrSize, StrComp
    Repeat
    Repeat <expr>
    Repeat While/Until <cond>
    Repeat <var> From <expr> To <expr>
    Repeat <var> From <expr> To <expr> step <expr>
    If / If-Else
    Case, single value and ranges
    Quit
    LookUp, LookUpZ, LookDown, LookDownZ, single and ranges
    ByteFill, WordFill, LongFill
    ByteMove, WordMove, LongMove
    WaitCnt, WaitPeq, WaitPne
    CogId, ChipVer
    Lock control
    Cog stops on end of program
    
  Not tested ...

    Clock setting
    WaitVid
    
  Opcodes not implemented ...
  
    15  MARK         
    28  COGIFUN 
    2C  COGISUB
    3C  ?
                                         
PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything along the
              lines of, "Based on original AiChip_SpinVm_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

 }
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE      = XTAL1 + PLL16x                        ' 80MHz System Clock Speed
  _XINFREQ      = 5_000_000

OBJ

  tv            : "TV_Text"                             ' TV Text Display used for debugging

CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  TV_PIN        = 12                                    ' Base TV Pin

' .-----------------------------------------------------------------------------------------------------.
' |     Serial Port Monitoring ( for HyperTerm etc )                                                    |
' `-----------------------------------------------------------------------------------------------------'

  RX_PIN        = -1                                    ' -Ve = None, 31 = PropPlug ( 8 = Used by Aichip )
  TX_PIN        = 30                                    ' -Ve = None, 30 = PropPlug ( 9 = Used by AiChip )

  BAUD_RATE     = 19200                                 ' Baud rate ( 19200 maximum )
  
VAR

  word infoBlock[8]                                     ' Info Block to launch Spin code
  long stack[1024]                                      ' Stack area to use
  long cogDump[$200]                                    ' Copy of cog memory when event occurs

' *******************************************************************************************************
' *                                                                                                     *
' *     Spin code interpreted by the Spin VM Interpreter, not the Parallax Spin Interpreter             *
' *                                                                                                     *
' *******************************************************************************************************

VAR

  long a
  word b
  byte c
  
PRI SpinCode ' No parameters and no local variables allowed here

  SpinSub

PRI SpinSub | localB

  repeat
    b := localB++
    a := ( c := ( b ) )
    case a
      -1 :
      other :
        case a
          3 : quit   
     
' *******************************************************************************************************
' *                                                                                                     *
' *     Spin code interpreted by the Parallax Spin Interpreter follows                                  *
' *                                                                                                     *
' *******************************************************************************************************
  
' There must only be one PUB in this source and that must be PUB Main and the PRI SpinCode must be the
' very first PRI in the source. This is because the Propeller Tool does not support @SpinCode so its
' address has to be hard-wired. Alter anything and it will all rapidly go bottom-up.

PRI Version
  return String("AiChip_SpinVm_004")

PUB Main

  tv.Start( TV_PIN )                                    ' Start the TV display running

  serial_Start                                          ' Start the serial port monitoring
  
  TxText( Version )

' StackTest
  
  infoBlock[ $0004 >> 1 ] := word[ $0004 ]              ' $0004 Clock Setting / Checksum
  infoBlock[ $0006 >> 1 ] := word[ $0006 ]              ' $0006 Object Base
  infoBlock[ $0008 >> 1 ] := word[ $0008 ]              ' $0008 Var Base
  infoBlock[ $000A >> 1 ] := @stack                     ' $000A Stack Base
  infoBlock[ $000C >> 1 ] := AtSpinCode                 ' $000C Initial PC
  infoBlock[ $000E >> 1 ] := @stack+4                   ' $000E Stack Top

  ' From here ( the Spin Program ), @SpinVm gives the absolute address of where the SpinVM is in
  ' hub memory. In the cog, 'ovlPtr long @SpinVm' holds the address of where SpinVM is relative to
  ' the start of the object the SpinVM is included in. The difference between the two gives a
  ' cog base pointer ( 'cogBp' )which needs to be added to any 'long @label' pointer used in the
  ' SpinVM to get an absolute hub address ( ie, in Overlay ). We set 'cogBp' before launching the
  ' SpinVm
  '
  ' The 'long @SpinVm' is placed in 'ovlPtr' in the cog so we can preset 'cogBp' to $0010 which is 
  ' the correct value to use when called from a top object ( main program ). Thus the value will
  ' quite often already be correctly set if this code is re-used and someone forgets to set 'cogBp'.

  if ChipVer <> 1
    TxText( String($D,$D,"SpinVM can currently only run on a Propeller Mk I") )
    repeat

  if ( @SpinVm - long[ @ovlPtr ] ) <> $0010
    TxText( String($D,$D,"SpinVM can currently only run from a top object") )
    repeat
     
  long[ @cogBp ] := @SpinVm - long[ @ovlPtr ]           ' Set the cog base pointer

  cognew( @SpinVm, @InfoBlock + $0004 )                 ' Launch the Spin VM

' TimeTest

  Debugger                                              ' Run the debugger

  ' The code never gets this far so the call to SpinCode never gets executed, just in case of 
  ' an editing error a repeat is thrown in to stop it going beyond that. The call to SpinCode 
  ' is simply so the AiChip bytecode decompiler can do a better job.

  repeat
  SpinCode
  
PRI Debugger | event, cogAdr
 
  repeat
    repeat while ( event := long[ $7FF0 ] ) == 0        ' Wait until Spin VM talks back
    TxByte( $0D )
    repeat cogAdr from $000 to $1FF
      long[ $7FF0 ] := 0
      repeat while long[ $7FF0] == 0
      cogDump[ cogAdr ] := long[ $7FEC ]
    case event
      1     : IsDebug(@Debug_Ret)                       ' 1 Debug
      2     : IsDebug(@Stop_Ret)                        ' 2 Stop
              TxText( String($0D,"Stop") )
              repeat
      3     : IsSingleStep                              ' 3 SingleStep
      4     : IsNotImplemented                          ' 4 NotImplemented
      other : TxText( String("Unknown ") )              ' ? Unknown
              TxHex( event, 8 )
              repeat
    long[ $7FEC ] := 0                                  ' Let the VM execute another instruction
    long[ $7FF0 ] := 0
   
PRI IsDebug( AtCogAddress )

  TxByte( "+")
  TxHex( ( Cog(AtCogAddress) & $1FF ) - 1, 3 )
  TxByte( " " )
  TxHex( long[ $7FF4 ], 8 )
  TxByte( " " )
  TxHex( long[ $7FF8 ], 8 )
  TxByte( " " )
  TxHex( long[ $7FFC ], 8 )
  
PRI IsNotImplemented

  TxText( String("Not Implemented before $") )
  TxHex( Cog(@pc), 4 )
  repeat
  
PRI IsSingleStep | xPc, xSp, xOp

  xPc := Cog(@pc)
  xSp := Cog(@sp)
  xOp := byte[ xPc ]
   
  TxHex( xSp, 4)
  TxByte( " " )
  TxHex( (xSp-@stack-4) >> 2, 2 )

  routeSerialOnly := true
  ShowStack( xSp, 4 )
  ShowStack( xSp, 3 )
  routeSerialOnly := false
   
  ShowStack( xSp, 2 )
  ShowStack( xSp, 1 )
  ShowStack( xSp, 0 )

  TxText( String(" a=") )
  TxDec( a )
  TxByte( $0D )
   
  TxHex( xPc, 4 )                                   ' Pc
  TxByte( " " )
  TxHex( xOp, 2 )                                   ' Opcode
  TxByte( " " )
   
  case xOp & %1110_0000
    %0000_0000, %0010_0000 :                        ' 00oo oooo
      ShowOpcode( xPc, xOp )
    %0100_0000, %0110_0000 :                        ' 01wv vvoo
      ShowFastAccess( xOp )
    %1000_0000, %1010_0000, %1100_0000 :            ' 1ssi wwoo
      ShowSlowAccess( xPc, xOp )
    %1110_0000 :                                    ' 111o oooo
      ShowOperator( xOp )

  PauseMs( 1000 )

PRI ShowStack( xSp, depth )

  TxByte( " " )
  if ((xSp-@stack-4) >> 2 ) > depth
    TxHex( long[ xSp - (depth*4) ], 8 )
  else
    TxText( String("--------") )
  
PRI Cog( AtCogAddress )
  result := cogDump[ ( AtCogAddress - @SpinVm ) >> 2 ]

PRI PauseMs( mS )
  waitcnt( CLKFREQ / 1000 * mS + CNT )
  
PRI ShowOpcode( xPc, xOp ) | i, xOpr

  i := 0
  case xOp
    $00 : i := String("FRAME 0  NO ABORT / RETVAL")
    $01 : i := String("FRAME 1  NO ABORT / NO RETVAL")
    $02 : i := String("FRAME 2  ABORT / RETVAL")
    $03 : i := String("FRAME 3  ABORT / NO RETVAL")
    $04 : i := String("GOTO")                            
    $05 : i := String("CALL")                            
    $06 : i := String("CALLOBJ")                         
    $07 : i := String("CALLOBJ  []")                     
    $08 : i := String("LOOPJPF")                         
    $09 : i := String("LOOPRPT")                         
    $0A : i := String("JPF")
    $0B : i := String("JPT")
    $0C : i := String("CASEJMP")
    $0D : i := String("CASE")
    $0E : i := String("CASERNG")
    $0F : i := String("LOOKEND")
    $10 : i := String("LOOKUP")
    $11 : i := String("LOOKDN")
    $12 : i := String("LOOKUPR")
    $13 : i := String("LOOKDNR")
    $14 : i := String("QUIT")
    $15 : i := String("MARK")
    $16 : i := String("STRSIZE")
    $17 : i := String("STRCOMP")
    $18 : i := String("BYTEFIL")
    $19 : i := String("WORDFIL")
    $1A : i := String("LONGFIL")
    $1B : i := String("WAITPEQ")
    $1C : i := String("BYTEMOV")
    $1D : i := String("WORDMOV")
    $1E : i := String("LONGMOV")
    $1F : i := String("WAITPNE")
    $20 : i := String("CLKSET")
    $21 : i := String("COGSTOP")
    $22 : i := String("LRETSUB")
    $23 : i := String("WAITCNT")
    $24 : i := String("PUSH     SPR[]")
    $25 : i := String("POP      SPR[]")
    $26 : i := String("USING    SPR[]")                
    $27 : i := String("WAITVID")
    $28 : i := String("COGIFUN")
    $29 : i := String("LNEWFUN")
    $2A : i := String("LSETFUN")
    $2B : i := String("LCLRFUN")
    $2C : i := String("COGISUB")
    $2D : i := String("LNEWSUB")
    $2E : i := String("LSETSUB")
    $2F : i := String("LCLRSUB")
    $30 : i := String("ABORT")
    $31 : i := String("ABOVAL")
    $32 : i := String("RETURN")
    $33 : i := String("RETVAL")
    $34, $35, $36 :
      TxText( String("PUSH     #") )
      TxDec( xOp - $35 )
    $37 :
      TxText( String("PUSH     #$") )
      xOp := byte[ xPc+1 ]
      xOpr := 2 << ( xOp & $1F )
      if xOp & $20
        xOpr--
      if xOp & $40
        xOpr ^= $FFFF_FFFF
      TxHex( xOpr, 8 )     
    $38, $39, $3A, $3B :
      TxText( String("PUSH     #$") )
      xOpr := 0
      repeat while --xOp => $37
        xOpr := ( xOpr << 8 ) | byte[ ++xPc ]
      TxHex( xOpr, 8 )     
    $3C : i := String("?")
    $3D, $3E, $3F :
      xOpr := byte[ xPc+1 ]
      if xOpr == $89
        i := String("PUSH   COGID")
      else
        case ( xOpr >> 5 ) & 3
          %00 : TxText( String("PUSH     ") )
          %01 : TxText( String("POP      ") )
          %10 : TxText( String("USING    ") )
          %11 : TxText( String("PUSH     #") )
        ShowSpr(xOpr)
        case xOp
          $3D : i := String("[]")
          $3E : i := String("[..]")
        if ( ( xOpr >> 5 ) & 3 ) == %10
          UsingOperator( xPc+2 )
  if i
    TxText( i )

PRI ShowFastAccess( xOp ) | xOpr

  xOpr := xOp & %0001_1100
  xOp  := ( ( xOp >> 3 ) & %0000_0100 ) | ( xOp & %0000_0011 ) | %1100_1000
  ShowAccess( xOp, xOpr )
  
PRI ShowSlowAccess( xPc, xOp ) | xOpr

  xOpr := byte[ xPc+1 ]
  if xOpr & %1000_0000
    xOpr := ( ( xOpr & %0111_1111 ) << 8 ) | byte[ xPc+2 ]
  ShowAccess( xOp, xOpr )

PRI ShowAccess( xOp, xOpr )

  case xOp & %0000_0011
    %0000_0000 : TxText( String("PUSH     ") )
    %0000_0001 : TxText( String("POP      ") )
    %0000_0010 : TxText( String("USING    ") )
    %0000_0011 : TxText( String("PUSH     #") )

  case xOp & %0000_1100
    %0000_0000 : TxText( String("MEM[]") )
    %0000_0100 : TxText( String("OBJ+") )
    %0000_1000 : TxText( String("VAR+") )
    %0000_1100 : TxText( String("LOC+") )

  if xOp & %0000_1100
    TxByte( "$" )
    TxHex( xOpr, 4 )

  if xOp & %0001_0000
    TxText( String("[]") )

  case xOp & %0110_0000
    %0000_0000 : TxText( String(".BYTE") )
    %0010_0000 : TxText( String(".WORD") )
    %0100_0000 : TxText( String(".LONG") )

PRI ShowOperator( xOp ) | i
  case xOp
    $E0 : i := String("ROR")
    $E1 : i := String("ROL")
    $E2 : i := String("SHR")
    $E3 : i := String("SHL")
    $E4 : i := String("HIGHEST")
    $E5 : i := String("LOWEST")
    $E6 : i := String("NEG")
    $E7 : i := String("BIT_NOT")
    $E8 : i := String("BIT_AND")
    $E9 : i := String("ABS")
    $EA : i := String("BIT_OR")
    $EB : i := String("BIT_XOR")
    $EC : i := String("ADD")
    $ED : i := String("SUB")
    $EE : i := String("SAR")
    $EF : i := String("BIT_REV")
    $F0 : i := String("LOG_AND")
    $F1 : i := String("BIT_GET")
    $F2 : i := String("LOG_OR")
    $F3 : i := String("BIT_SET")
    $F4 : i := String("MPY")
    $F5 : i := String("MPY_BIG")
    $F6 : i := String("DIV")
    $F7 : i := String("MOD")
    $F8 : i := String("SQRT")
    $F9 : i := String("LT")
    $FA : i := String("GT")
    $FB : i := String("NE")
    $FC : i := String("EQ")
    $FD : i := String("LE")
    $FE : i := String("GE")
    $FF : i := String("LOG_NOT")
  TxText( i )

PRI ShowSpr( sprNum )
  case sprNum & $F
    $0 : TxText( String("PAR") )
    $1 : TxText( String("CNT") )
    $2 : TxText( String("INA") )
    $3 : TxText( String("INB") )
    $4 : TxText( String("OUTA") )
    $5 : TxText( String("OUTB") )
    $6 : TxText( String("DIRA") )
    $7 : TxText( String("DIRB") )
    $8 : TxText( String("CTRA") )
    $9 : TxText( String("CTRB") )
    $A : TxText( String("FRQA") )
    $B : TxText( String("FRQB") )
    $C : TxText( String("PHSA") )
    $D : TxText( String("PHSB") )
    $E : TxText( String("VCFG") )
    $F : TxText( String("VSCL") )

PRI UsingOperator( xPc ) | xOp, i

  TxByte( " " )
  xOp := byte[xPc]
  if xOp & $80
    TxText( String("PUSH ") )
    xOp &= $7F
  if xOp => $40
    ShowOperator(xOp | $E0 )
  else
    i := String("?")
    case xOp
      $00 : i := String("PREPOP")
      $02 : i := String("RPTINCJ")
      $06 : i := String("RPTADDJ")
      $08 : i := String("FWDRAND")
      $0C : i := String("REVRAND")
      $10 : i := String("PRESEXB")
      $14 : i := String("PRESEXW")
      $18 : i := String("POSTCLR")
      $1C : i := String("PSOTSET")
      $22 : i := String("PREINC.BYTE")
      $24 : i := String("PREINC.WORD")
      $26 : i := String("PREINC.LONG")
      $2A : i := String("POSTINC.BYTE")
      $2C : i := String("POSTINC.WORD")
      $2E : i := String("POSTINC.LONG")
      $32 : i := String("PREDEC.BYTE")
      $34 : i := String("PREDEC.WORD")
      $36 : i := String("PREDEC.LONG")
      $3A : i := String("POSTDEC.BYTE")
      $3C : i := String("POSTDEC.WORD")
      $3E : i := String("POSTDEC.LONG")
    TxText(i)
                                                                      
PRI IsRcscSingleStep | xPc, xSp, xOp

  xPc := Cog(@pc)
  xSp := Cog(@sp)
  xOp := byte[ xPc ]

  TxText( String("RCSC ") )
  if xPc < 0
    TxText( String("Initialisation ") )
    TxHex( xPc, 8)
    TxByte( " " )
    TxHex( -xPc, 4 )       
  else
    TxHex( xPc, 4)
    TxByte( " " )
    TxHex( xOp, 2 )        

  PauseMs(1000)
  
PRI AtSpinCode                                          ' Faked result := @SpinCode
  result := word[ $0006 ]
  result := word[ result ][4] + result

CON

' *******************************************************************************************************
' *                                                                                                     *
' *     Display Handler                                                                                 *
' *                                                                                                     *
' *******************************************************************************************************

VAR

  long txInvert
  long rxInvert
  long bitTime

  long numberRadix
  long numberWidth

  long routeSerialOnly
  
PRI TxByte( n )

  if TV_PIN => 0
    if not routeSerialOnly
      tv.Out( n )
    
  if TX_PIN => 0
    serial_Out( n )
    if n == $0D
      serial_Out( $0A )
    
PRI TxText( txtPtr )

  repeat strsize(txtPtr)
    TxByte(byte[txtPtr++])

PRI TxDec( n )

  TxNumberFormat( n, 10, 0 )
  
PRI TxHex( n, w )

  TxNumberFormat( n, 16, w )

PRI TxBin( n, w )

  TxNumberFormat( n, 2, w )
  
PRI TxNumberFormat( n, radix, width )

  numberRadix := radix
  numberWidth := width
  TxNumber(n)

PRI TxNumber( n ) | radix, digits, k

  radix := numberRadix

  if n == 0 
    repeat numberWidth #> 1
      TxByte("0")
  else
   
    case radix
   
      2  : TxRadixNumber(n,31,$1,1,32)
      4  : TxRadixNumber(n,30,$3,2,16)
      8  : TxRadixNumber(n,30,$7,3,11)
      16 : TxRadixNumber(n,28,$F,4,8)
   
      other : k := 1_000_000_000
              digits := 11
              if n < 0
                TxByte("-")
                n := -n
                digits--
              repeat while k > n
                k /= 10
                digits--
              repeat while numberWidth => digits++
                TxByte("0")
              repeat while k <> 0
                TxDigit( ( n / k ) // 10 )
                k /= 10

PRI TxRadixNumber(n,shift,mask,decr,digits)

  repeat while ( n >> shift ) == 0
    shift -= decr
    digits--
  repeat while numberWidth > digits++
    TxByte("0")
  repeat while shift => 0
    TxDigit( ( n >> shift ) & mask )
    shift -= decr

PRI TxDigit( n )

  if ( n += "0" ) > "9"
    n := n - "9" + "A" - 1
  TxByte( n )

' *******************************************************************************************************
' *                                                                                                     *
' *     Serial Interfacing                                                                              *
' *                                                                                                     *
' *******************************************************************************************************

PRI Serial_Start

  if TX_PIN => 0
    txInvert := ( TX_PIN => 28 )
    outa[TX_PIN] := txInvert
    dira[TX_PIN] := 1

  if RX_PIN => 0
    rxInvert := ( RX_PIN => 28 )
    dira[RX_PIN] := 0

  bitTime := clkfreq / ( BAUD_RATE <# 19200 )
   
PRI Serial_Out( n ) | timeout

  n := (( (( n & $FF ) ^$FF ) << 2 ) | 2 ) ^ txInvert
  timeout := cnt
  repeat 11
    waitcnt(timeout += bitTime)
    outa[TX_PIN] := ( n >>= 1 ) & 1

CON

  OVERLAY_SIZE   = 26                                   ' How large the overlay area is
  REGISTERS_SIZE = 3                                    ' How mny local registers for the overlay

  LMM            = 1                                    ' Flag to indicate LMM-style overlays
                                                        
DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Spin VM Interpreter                                                                             *
' *                                                                                                     *
' *******************************************************************************************************
'
SpinVm          org     $000

                jmp     #OVERLAY_START                  ' Initialise VM
                
' *******************************************************************************************************
' *                                                                                                     *
' *     Debugging Utilities                                                                             *
' *                                                                                                     *
' *******************************************************************************************************
'
' Debugging uses the top 20 bytes ( 5 longs ) of Ram ( $7FEC to $7FFF ) to communicate with the Spin
' wrapper.
 
' These routines allow the cog to inform the Spin wrapper what's going on within itself. In all cases 
' the long[$7FF0] is set with a flag to indicate what internal 'event' is being generated. Then, once
' long[$7FF0] has been cleared, the entire $200 longs of cog are passed to the Spin program a long at 
' a time. This is done by putting the cog value in long[$7FEC] and setting long[$7FF0]. Every time the
' long[$7FF0] is cleared by the Spin wrapper the next long is passed out. After $200 have been passed
' over the code here waits for long[$7FF0] to be cleared again and eecution continues. It is up to the
' Spin wrapper to not allow the cog to continue after a 'Stop' or 'NotImplemented' event.
'
' Note that during the copy we load 'dbgTmp' then wrlong that rather than altering the destination of
' the wrlong itself; this is so we can read the SPR's which cannot be used as destination registers. It
' also helps with the counting method used.
'
' In addition, writes to long[$7FF4], long[$7FF8] and long[$7FFC] can be used to pass information back
' to the debugger prior to calling any of the debugging routines. These are displayed on Debug and Stop
' calls. The Spin wrapper dictates how the events are handled and what is displayed.
'
' Example ...
'
'               wrlong  lhs,k_0000_7FF4
'               wrlong  rhs,k_0000_7FF8
'               wrlong  acc,k_0000_7FFC
'               call    #Debug
'
' Note we have to protect the Z flag so we don't screw any Cog code which relies on it being unchanged
' over calls to Debug etc. We do that by using the lsb of the destination register of the Debug_Ret
' which is otherwise unused. It saves a long. We don't need to protect C because we don't use WC in the
' debugger. 

NotImplemented  add    dbgTmp,#1                        ' 4
SingleStep      add    dbgTmp,#1                        ' 3
Stop            add    dbgTmp,#1                        ' 2
Debug           add    dbgTmp,#1                        ' 1

                wrlong  dbgTmp,k_0000_7FF0

                muxnz   Debug_Ret,k_0000_0200           ' Protect Z flag

:Dump           call    #DebugWait
:Src            mov     dbgTmp,$000                     ' No need to pre-initialise this
                wrlong  dbgTmp,k_0000_7FEC
                add     :Src,#1
                and     :Src,#$1FF WZ, NR               ' If we hit $200 we corrupted the opcode
        IF_Z    sub     :Src,k_0000_0200                ' So set it to $000 ...
                wrlong  k_FFFF_FFFF,k_0000_7FF0
        IF_NZ   jmp     #:Dump                          ' ... and no more to send out

                call    #DebugWait                      ' leaves dbgTmp zeroed on exit

                and     Debug_Ret,k_0000_0200 WZ, NR    ' Restore Z flag

NotImplemented_Ret
SingleStep_Ret
Stop_Ret
Debug_Ret       ret

DebugWait       rdlong  dbgTmp,k_0000_7FF0
                tjnz    dbgTmp,#DebugWait
DebugWait_Ret   ret

dbgTmp          long    0

k_0000_7FEC     long    $0000_7FEC
k_0000_7FF0     long    $0000_7FF0
k_0000_7FF4     long    $0000_7FF4
k_0000_7FF8     long    $0000_7FF8
k_0000_7FFC     long    $0000_7FFC

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Instruction Fetch and Dispatch                                                                  *
' *                                                                                                     *
' *******************************************************************************************************
'
'               .---.---.---.---.---.---.---.---.
'               | 0 | 0 | o   o   o   o   o   o |         Command Opcodes
'               `---^---^---^---^---^---^---^---'
'               .---.---.---.---.---.---.---.---.
'               | 0   1 | w | v   v   v | o   o |         Fast Access Opcodes
'               `---^---^---^---^---^---^---^---'
'               .---.---.---.---.---.---.---.---.
'               | 1 | s   s | i | w   w | o   o |         Slow Access Opcodes
'               `---^---^---^---^---^---^---^---'
'               .---.---.---.---.---.---.---.---.
'               | 1   1   1 | o   o   o   o   o |         Operators
'               `---^---^---^---^---^---^---^---'

Fetch           call    #SingleStep

                rdbyte  opc,pc                          ' Get opcode
                add     pc,#1

                and     opc,#%1000_0000 WZ, NR
        IF_NZ   jmp     #SlowAccess

                and     opc,#%0100_0000 WZ, NR
        IF_NZ   jmp     #FastAccess

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Command Opcode Dispatch                                                                         *
' *                                                                                                     *
' *******************************************************************************************************
'
'               .---.---.---.---.---.---.---.---.
'               | 0 | 0 | o   o   o   o   o   o |
'               `---^---^---^---^---^---^---^---'
'                
'               00  FRAME        | 10  LOOKUP       | 20  CLKSET       | 30  ABORT
'               01  FRAME        | 11  LOOKDN       | 21  COGSTOP      | 31  ABOVAL
'               02  FRAME        | 12  LOOKUPR      | 22  LRETSUB      | 32  RETURN
'               03  FRAME        | 13  LOOKDNR      | 23  WAITCNT      | 33  RETVAL
'               04  GOTO         | 14  QUIT         | 24  PUSH   SPR[] | 34  PUSH   #-1
'               05  CALL         | 15  MARK         | 25  POP    SPR[] | 35  PUSH   #0
'               06  CALLOBJ      | 16  STRSIZE      | 26  USING  SPR[] | 36  PUSH   #1
'               07  CALLOBJ[]    | 17  STRCOMP      | 27  WAITVID      | 37  PUSH   #kp
'               08  LOOPJPF      | 18  BYTEFIL      | 28  COGIFUN      | 38  PUSH   #k1
'               09  LOOPRPT      | 19  WORDFIL      | 29  LNEWFUN      | 39  PUSH   #k2
'               0A  JPF          | 1A  LONGFIL      | 2A  LSETFUN      | 3A  PUSH   #k3
'               0B  JPT          | 1B  WAITPEQ      | 2B  LCLRFUN      | 3B  PUSH   #k4
'               0C  GOTO[]       | 1C  BYTEMOV      | 2C  COGISUB      | 3C  ?
'               0D  CASE         | 1D  WORDMOV      | 2D  LNEWSUB      | 3D  REG    r[]
'               0E  CASERNG      | 1E  LONGMOV      | 2E  LSETSUB      | 3E  REG    r[..]
'               0F  LOOKEND      | 1F  WAITPNE      | 2F  LCLRSUB      | 3F  REG    r

Opcode          mov     ovlPtr,opc                      ' Setup indexing for the lookup table
                shr     ovlPtr,#1                       ' Two entries in per lookup
                add     ovlPtr,#Opcode_Table            ' Index into the table
                movs    Opcode_Src,ovlPtr               ' Set the index ready for reading

                shr     opc,#1 WC, NR                   ' Determine which half of the lookup to use
              
Opcode_Src      mov     ovlPtr,0-0                      ' Get the table entry

        IF_C    shr     ovlPtr,#16                      ' Use the MSW
                and     ovlPtr,k_0000_FFFF              ' And only use a 16-bit address

                movs    CommonReturn,#Fetch             ' All returns from overlay go to Fetch

' *******************************************************************************************************
' *                                                                                                     *
' *     Overlay Loader                                                                                  *
' *                                                                                                     *
' *******************************************************************************************************
'
' The overlay loader is called by jumping to 'Overlay_This' with 'ovlPtr' having been loaded with the
' Hub address of where to load a block of code from. Up to OVERLAY_SIZE longs of hub memory are loaded
' between OVERLAY_START and OVERLAY_END-1 of Cog. If the block being loaded contains a 'long 0', nothing
'  after that will be loaded. This is a convenient trick to speed up the overlay loading. The 'long 0'
' is automatically replaced by an effective 'jmp #CommonReturn' ( the actual CommonReturn instruction
' is copied down ).
'
' The Overlay routine is used as general purpose opcode dispatcher, and when 'Overlay_This' is reached,
' 'ovlPtr' will containe one of three things -
'
'       A cog addrees of an internal routine            value 0000_0000 to 0000_01EF
'       A hub address of overlay to load                value 0000_01F0 to 0000_FFFF
'       A PASM instruction                              value XXXX_0000 to XXXX_FFFF ( X is non-zero )
'
' Overlays are always placed after the Spin VM code which is forced to be at least $1F0/4 longs in length
' so internal routines to execute can be distinguished from external hub addresses. PASM instructions
' will have at least one bit of the conditional ( IF_XX ) execution bits set so they can be distinguished
' as well. When a Hub address is provided an overlay is loaded as previously described. When an internal
' address is provided that is simply jumped to. When a PASM instruction is provided, that is placed at
' the last long of the overlay and jumped to, the 'CommonReturn' immediately follows so control will be
' passed back to wherever it was invoked from.
                
'               At this point, 'lhs' is a hub address,
'               a cog adress or a PASM instruction

Overlay_This    cmp     ovlPtr,#$1F0 WZ, WC, NR         ' If the address is less than $1F0 it is a cog
        IF_B    jmp     ovlPtr                          ' address and we simply jump to it

'               It's an overlay so read it in or start
'               interpreting it
'
'               Need to adjust the ovlPtr because all
'               @label values are relative to object
'               base and not actual hub addresses               

                add     ovlPtr,cogBp                    ' Add Cog Base Pointer

                shr     ovlPtr,#1 WC, NR                ' Is it an LMM overlay ?
        IF_NC   jmp     #:Load                          ' No - Load and execute it

'               Do an LMM interpretation. We allow the
'               'long 0' to execute as a 'nop' because
'               that's better  than checking for it
'               before executing as that screws the 
'               hub access timing badly.

:Lmm            rdlong  :Opc,ovlPtr                     ' Read an instruction
                add     ovlPtr,#4
:Opc            nop                                     ' Execute the instruction
                tjnz    :Opc,#:Lmm                      ' Wasn't a 'long 0' so keep going
                
                jmp     #CommonReturn                   ' Finished

'               Load and execute an overlay

:Load           movd    :Dst1,#OVERLAY_START            ' Start copying to OVERLAY_START
                mov     ovlCnt,#OVERLAY_SIZE            ' And do not over-fill the overlay

'               Do the actual overlay loading

:Dst1           rdlong  0-0,ovlPtr WZ                   ' Read a long from hub to cog
        IF_NZ   add     :Dst1,k_0000_0200               ' Update the cog destination address
        IF_NZ   add     ovlPtr,#4                       ' Update the hub source address
        IF_NZ   djnz    ovlCnt,#:Dst1                   ' Keep going if there's more to copy

'               If we terminated on a 'long 0' turn
'               that into a 'jmp #CommonReturn'

        IF_NZ   jmp     #OVERLAY_START                  ' We did not so Pass control to the overlay
        
                mov     ovlPtr,:Dst1                    ' Get last cog address written
                shr     ovlPtr,#9
                movd    :Dst2,ovlPtr
                nop
:Dst2           mov     0-0,CommonReturn                ' Set the Common return then fall through ...
                
' .-----------------------------------------------------------------------------------------------------.
' |     Initialisation                                                                                  |
' `-----------------------------------------------------------------------------------------------------'

' Stack on exit ...                       Or if any local vars used ...
'                                           
'               .------.------.                         .------.------.
'     BP -----> | FFF9 | FFFF |               BP -----> | FFF9 | FFFF |
'               |------^------|                         |------^------|
'     SP -----> | 0000  00000 |                         | 0000  00000 |
'               `-------------'                         |-------------|
'                                                       | local var 1 |         Not zeroed
'                                                       |-------------|
'                                             SP -----> | local var N |         Not zeroed
'                                                       `-------------'
'
' Note that we write long $FFFF_FFF9 to bp which works because we should be writing two separate words
' and this sorts out the endianess problem.

   
OVERLAY_START   mov     adr,PAR                         ' 0004 CLK | CHK - We ignore these
                add     adr,#2
                rdword  op,adr                          ' 0006 Object Base
                add     adr,#2
                rdword  vp,adr                          ' 0008 Var Base
                add     adr,#2
                rdword  bp,adr                          ' 000A Stack Base
                add     adr,#2
                rdword  pc,adr                          ' 000C Initial PC
                add     adr,#2
                rdword  sp,adr                          ' 000E Initial SP
                
                mov     acc,:k_FFFF_FFF9                ' Init base pointer contents
                wrlong  acc,bp

                add     bp,#4                           ' Init 'result' variable
                wrlong  k_0000_0000,bp
                sub     bp,#4
                
                jmp     #Fetch

:k_FFFF_FFF9    long    $FFFF_FFF9

                long    0[OVERLAY_START+OVERLAY_SIZE-$] ' Pad out the overlay
       
OVERLAY_END     jmp     #OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     Common Return Point                                                                             |
' `-----------------------------------------------------------------------------------------------------'
'
' The common return point is placed immediately after the overlay so an overlay can 'fall-through' to
' return to where it was called from and so that PASM instructions can be placed in the last long of
' the overlay, execute and then have somewhere to go after that.

                org     OVERLAY_END
                
CommonReturn    res     1                               ' Yes, this must be a 'res 1'

' .-----------------------------------------------------------------------------------------------------.
' |     Command Opcode Jump table                                                                       |
' `-----------------------------------------------------------------------------------------------------'
'
' These are the overlay vectors to each of the handlers for the 64 Command Opcodes. They are packed as 
' two words per long to reduce the table size to 32 longs. 
'
' Each entry in this table can be one of three things ...
'
' 1) A 'word OpcLabel' which is the cog address of a handler within the cog.
'
' 2) A 'word @OvcLabel' which is the hub address of an overlay handler which will be loaded into the
'    cog and subsequently executed.
'
' 3) A 'word @OvcLabel | LMM' which is the hub address of an overlay handler which will be interpreted
'    Large Memory Model style.
'
' Overlays interpreted LMM-style cannot contain loops ( jumps to within themselves ), self-modifying
' code nor any local constants. Including any of those will lead to failure.

Opcode_Table    word    OpcFRAME                        ' $00   FRAME   0
                word    OpcFRAME                        ' $01   FRAME   1
                word    OpcFRAME                        ' $02   FRAME   2
                word    OpcFRAME                        ' $03   FRAME   3
                word    @OvcGOTO     | LMM              ' $04   GOTO
                word    OpcCALL                         ' $05   CALL
                word    @OvcCALLOBJ  | LMM              ' $06   CALLOBJ
                word    @OvcCALLOBJ  | LMM              ' $07   CALLOBJ []
                word    OpcLOOPXXX                      ' $08   LOOPJPF
                word    OpcLOOPXXX                      ' $09   LOOPRPT
                word    @OvcJPX      | LMM              ' $0A   JPF
                word    @OvcJPX      | LMM              ' $0B   JPT
                word    @OvcCASEJMP  | LMM              ' $0C   CASEJMP
                word    @OvcCASE     | LMM              ' $0D   CASE
                word    @OvcCASE     | LMM              ' $0E   CASERNG
                word    @OvcLOOKEND  | LMM              ' $0F   LOOKEND
                word    @OvcLOOKUPX                     ' $10   LOOKUP
                word    @OvcLOOKDNX                     ' $11   LOOKDN
                word    @OvcLOOKUPX                     ' $12   LOOKUPR
                word    @OvcLOOKDNX                     ' $13   LOOKDNR
                word    @OvcQUIT     | LMM              ' $14   QUIT
                word    @OvcMARK     | LMM              ' $15   MARK
                word    @OvcSTRSIZE                     ' $16   STRSIZE
                word    @OvcSTRCOMP                     ' $17   STRCOMP
                word    @OvcXXXXFIL                     ' $18   BYTEFIL
                word    @OvcXXXXFIL                     ' $19   WORDFIL
                word    @OvcXXXXFIL                     ' $1A   LONGFIL
                word    @OvcWAITPXX  | LMM              ' $1B   WAITPEQ
                word    @OvcXXXXMOV                     ' $1C   BYTEMOV
                word    @OvcXXXXMOV                     ' $1D   WORDMOV
                word    @OvcXXXXMOV                     ' $1E   LONGMOV
                word    @OvcWAITPXX  | LMM              ' $1F   WAITPNE
                word    @OvcCLKSET   | LMM              ' $20   CLKSET
                word    @OvcCOGSTOP  | LMM              ' $21   COGSTOP
                word    @OvcLRETSUB  | LMM              ' $22   LRETSUB
                word    @OvcWAITCNT  | LMM              ' $23   WAITCNT
                word    @OvcXXX_SPR  | LMM              ' $24   PUSH    SPR[]
                word    @OvcXXX_SPR  | LMM              ' $25   POP     SPR[]
                word    @OvcXXX_SPR  | LMM              ' $26   USING   SPR[]
                word    @OvcWAITVID  | LMM              ' $27   WAITVID
                word    @OvcCOGIXXX  | LMM              ' $28   COGIFUN
                word    @OvcLNEWXXX  | LMM              ' $29   LNEWFUN
                word    @OvcLXXXXXX  | LMM              ' $2A   LSETFUN
                word    @OvcLXXXXXX  | LMM              ' $2B   LCLRFUN
                word    @OvcCOGIXXX  | LMM              ' $2C   COGISUB
                word    @OvcLNEWXXX  | LMM              ' $2D   LNEWSUB
                word    @OvcLXXXXXX  | LMM              ' $2E   LSETSUB
                word    @OvcLXXXXXX  | LMM              ' $2F   LCLRSUB
                word    OpcRETURN                       ' $30   ABORT
                word    OpcRETURN                       ' $31   ABOVAL
                word    OpcRETURN                       ' $32   RETURN
                word    OpcRETURN                       ' $33   RETVAL
                word    OpcPUSH_K                       ' $34   PUSH    #-1
                word    OpcPUSH_K                       ' $35   PUSH    #0
                word    OpcPUSH_K                       ' $36   PUSH    #1
                word    OpcPUSH_Kp                      ' $37   PUSH    #kp
                word    OpcPUSH_Kn                      ' $38   PUSH    #k1
                word    OpcPUSH_Kn                      ' $39   PUSH    #k2
                word    OpcPUSH_Kn                      ' $3A   PUSH    #k3
                word    OpcPUSH_Kn                      ' $3B   PUSH    #k4
                word    @OvcUNKNOWN                     ' $3C   ?
                word    @OvcREG_LMM  | LMM | 1          ' $3D   REG     r[]
                word    @OvcREG_LMM  | LMM | 1          ' $3E   REG     r[..]
                word    @OvcREG_LMM  | LMM | 1          ' $3F   REG     r

DAT
                
' *******************************************************************************************************
' *                                                                                                     *
' *     Fast Access Dispatch                                                                            *
' *                                                                                                     *
' *******************************************************************************************************
'
' Fast Access opcodes are the same as Slow Access opcodes but are only one byte long and contain the
' address of the related operand within them. These are used to access the first 8 longs of the local
' stack or global vars for an object.
'
' Handling a Fast Access opcode is simply a case of converting the Fast Access Opcode into an expanded
' Slow Access opcode and leaving the Access handler to do its job.
'
'                         .------------------------------ 0 = VAR
'                         |                               1 = LOC
'                         |
'                         |       .---------------------- Address ( adr = v * 4 )
'                         |       |
'                         |       |         .------------ 0 0 = PUSH
'                         |       |         |             0 1 = POP
'                         |       |         |             1 0 = USING
'                         |       |         |             1 1 = PUSH #
'                         |       |         |
'               .---.---.---.---.---.---.---.---.
'               | 0   1 | w | v   v   v | o   o |
'               `---^---^---^---^---^---^---^---'
'                         |       |         |
'                         |       `---------|------------------------.
'                         |                 |                        |
'                         `-----------.     |                        |
'                                    \|/   \|/                      \|/
'               .---.---.---.---.---.---.---.---.  .---.---.---.---.---.---.---.---.
'               | 1 | 1   0 | 0 | 1   w | o   o |  | 0   0   0 | v   v   v | 0   0 |
'               `---^---^---^---^---^---^---^---'  `---^---^---^---^---^---^---^---'
'                       |           |
'                       |           `-------------------- 1 0 = VAR
'                       |                                 1 1 = LOC
'                       |
'                       `-------------------------------- 1 0 = Long

FastAccess      mov     opr,opc
                and     opr,#%0001_1100

                and     opc,#%0010_0000 WZ, NR
                muxnz   opc,#%0000_0100
                and     opc,#%0000_0111
                or      opc,#%1100_1000

                jmp     #Access

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Access Dispatch                                                                                 *
' *                                                                                                     *
' *******************************************************************************************************
'
'                       .-------------------------------- 0 0 = Byte
'                       |                                 0 1 = Word
'                       |                                 1 0 = Long
'                       |                                 1 1 = Unused - Unary / Binary operators
'                       |
'                       |     .-------------------------- 1 = []
'                       |     |
'                       |     |     .-------------------- 0 0 = MEM[]
'                       |     |     |                     0 1 = OBJ
'                       |     |     |                     1 0 = VAR
'                       |     |     |                     1 1 = LOC
'                       |     |     |
'                       |     |     |       .------------ 0 0 = PUSH
'                       |     |     |       |             0 1 = POP
'                       |     |     |       |             1 0 = USING
'                       |     |     |       |             1 1 = PUSH #
'                       |     |     |       |
'               .---.---.---.---.---.---.---.---.
'               | 1 | s   s | i | w   w | o   o |
'               `---^---^---^---^---^---^---^---'

SlowAccess      cmp     opc,#%1110_0000 WC, WZ, NR
        IF_AE   jmp     #Operator
        
                and     opc,#%0000_1100 WZ, NR          ' Is it MEM ?
        IF_Z    call    #PopLhs                         ' Yes, no operand used, operand on stack
        IF_Z    mov     opr,lhs
        IF_Z    jmp     #Access
        
                rdbyte  opr,pc                          ' Get operand address
                add     pc,#1
                and     opr,#%1000_0000 WZ, NR          ' 0aaa aaaa
                and     opr,#%0111_1111                 ' 1aaa aaaa aaaa aaaa
        IF_NZ   shl     opr,#8
        IF_NZ   rdbyte  rhs,pc
        IF_NZ   add     pc,#1
        IF_NZ   or      opr,rhs

Access          mov     rhs,opc                         ' 1ssi wwoo
                and     rhs,#%0000_1100                 '      ZC
                shr     rhs,#2

                mov     acc,#1                          ' Determine address size stepping
                shl     acc,rhs

                and     opc,#%0001_0000 WZ, NR          ' [] Array indexing
  IF_NZ         mov     acc,opc                         ' Determine size of entity
  IF_NZ         and     acc,#%0110_0000
  IF_NZ         shr     acc,#5
  IF_NZ         shl     opr,acc                         ' Adjust index for size of entity
  IF_NZ         call    #PopLhs                         
  IF_NZ         add     opr,lhs

                shr     rhs,#1 WC, WZ, NR               ' 00 MEM[]
  IF_Z_AND_C    add     opr,op                          ' 01 OBJ
  IF_NZ_AND_NC  add     opr,vp                          ' 10 VAR
  IF_NZ_AND_C   add     opr,bp                          ' 11 LOC
  IF_NZ_AND_C   add     opr,#4                          ' 11 LOC - BP is one long below the result (LOC+0)

Access_Spr_Here mov     rhs,opc                         ' 1ssi wwoo
                and     rhs,#%0110_0000                 '  ZC
                shr     rhs,#5
                shr     rhs,#1 WC, WZ
     
                and     opc,#%0000_0011                 ' 1ssi wwoo
                xor     opc,#%0000_0001                 ' This allows %10 USING to fall through
                shl     opc,#1
                add     opc,#Access_Jump
                jmp     opc

Access_Jump     call    #PopThenPut                     ' 00 <== 01 POP
                jmp     #Fetch

                call    #GetThenPush                    ' 01 <== 00 PUSH
                jmp     #Fetch

                mov     lhs,opr                         ' 10 <== 11 PUSH #
                jmp     #Push

             '  jmp     #Using                          ' 11 <== 10 USING , falls through to next ...

' .-----------------------------------------------------------------------------------------------------.
' |     Handle USING opcodes                                                                            |
' `-----------------------------------------------------------------------------------------------------'
'
'                 .-------------------------------------- 1 = Push Result
'                 |       
'               .---.---.---.---.---.---.---.---.
'               | p | 0   0 | o   o   o   o | 0 |         Special Operators
'               `---^---^---^---^---^---^---^---'
'               .---.---.---.---.---.---.---.---.
'               | p | 0   1 | o   o   o   o | 0 |         Pre/Post Inc/Dec
'               `---^---^---^---^---^---^---^---'
'               .---.---.---.---.---.---.---.---.
'               | p | 1 | o   o   o   o   o   o |         Operators
'               `---^---^---^---^---^---^---^---'
'
'               00  PREPOP       | 10  PRESEXB      | 20               | 30
'               02  RPTINCJ      | 12               | 22  PREINC Byte  | 32  PREDEC  Byte
'               04               | 14  PRESEXW      | 24  PREINC Word  | 34  PREDEC  Word
'               06  RPTADDJ      | 16               | 26  PREINC Long  | 36  PREDEC  Long
'               08  FWDRAND      | 18  POSTCLR      | 28               | 38  
'               0A               | 1A               | 2A  POSTINC Byte | 3A  POSTDEC Byte
'               0C  REVRAND      | 1C  POSTSET      | 2C  POSTINC Word | 3C  POSTDEC Word
'               0E               | 1E               | 2E  POSTINC Long | 3E  POSTDEC Long
                                                                      
'               40  ROR          | 58  BIT_AND      | 60  LOG_AND      | 78  SQRT
'               41  ROL          | 59  ABS          | 61  BIT_GET      | 79  LT
'               42  SHR          | 5A  BIT_OR       | 62  LOG_OR       | 7A  GT
'               43  SHL          | 5B  BIT_XOR      | 63  BIT_SET      | 7B  NE
'               44  HIGHEST      | 5C  ADD          | 64  MPY          | 7C  EQ
'               45  LOWEST       | 5D  SUB          | 65  MPY_BIG      | 7D  LE
'               46  NEG          | 5E  SAR          | 66  DIV          | 7E  GE
'               47  BIT_NOT      | 5F  BIT_REV      | 67  MOD          | 7F  LOG_NOT

Using           mov     flags,#0                        ' Remember flag settings for operand size
                muxc    flags,#1                        
                muxnz   flags,#4                        

                rdbyte  opc,pc                          ' Get the effects to apply
                add     pc,#1
 
                and     opc,#%1000_0000 WZ, NR          ' Do we have a push ?
                muxnz   flags,#2                        ' Set bit2 if yes                        

'               .---------------------------------------------------------------------------------------.
'               | $40 to $7F                                                                            |
'               `---------------------------------------------------------------------------------------'

                and     opc,#%0100_0000 WZ, NR          ' Is it standard operator ( eg += ) ?
        IF_Z    jmp     #Using_1                        ' No

                movs    Opr_Src,opc                     ' Yes
                andn    Opr_Src,#%1100_0000
                mov     opcBit,#1                       ' Turn opcode into an opcode bit
                shl     opcBit,Opr_Src
                add     Opr_Src,#Operator_Table         

                jmp     #Using_2                        ' Get the operand and process it

'               .---------------------------------------------------------------------------------------.
'               | $00 to $1F                                                                            |
'               `---------------------------------------------------------------------------------------'

Using_1         mov     opcBit,#0                       ' Never binary, never boolean

                and     opc,#%0010_0000 WZ, NR          ' Is it Pre/Post Inc/dec ?
        IF_NZ   jmp     #Using_2                        ' Yes

                mov     acc,opc                         ' No, Special case ( eg a~~ )
                shr     acc,#2
                movs    Opr_Src,acc                     
                andn    Opr_Src,#%1111_1000
                and     Opc,#%0111_1111 WZ, NR          ' Relocate PREPOP
        IF_Z    or      Opr_Src,#%0000_1000
                add     Opr_Src,#Using_Table

'               .---------------------------------------------------------------------------------------.
'               | Get the operand and process it                                                        |
'               `---------------------------------------------------------------------------------------'

Using_2         shr     flags,#2 WC, WZ, NR             ' Recall size needed

                call    #Get                            ' Get lhs

                mov     rhs,lhs                         ' Put in rhs as well
                
                and     opcBit,k_BinaryOps WZ, NR       ' Is it a binary op ?
        IF_NZ   rdlong  rhs,sp                          ' Yes, Pop rhs
        IF_NZ   sub     sp,#4                           

                and     opc,#%0010_0000 WZ, NR          ' Is it Pre/Post Inc?dec ?
        IF_NZ   jmp     #Using_3                        ' Yes

                jmpret  CommonReturn,#Opr_DoIt          ' Handle the operator

                shr     flags,#2 WC, WZ, NR             ' Recall size needed
                jmp     #Using_4                        ' Put the result back

'               .---------------------------------------------------------------------------------------.
'               | $20 to $3F - Pre/Post Inc/Dec                                                         |
'               `---------------------------------------------------------------------------------------'
'
'               20               | 30
'               22  PREINC Byte  | 32  PREDEC  Byte
'               24  PREINC Word  | 34  PREDEC  Word
'               26  PREINC Long  | 36  PREDEC  Long
'               28               | 38  
'               2A  POSTINC Byte | 3A  POSTDEC Byte
'               2C  POSTINC Word | 3C  POSTDEC Word
'               2E  POSTINC Long | 3E  POSTDEC Long
'
'                  .----------- 0 = Inc
'                  |            1 = Dec
'                  |
'                  | .--------- 0 = Pre
'                  | |          1 = Post
'                  | |
'               0010 0xx0       PRE  INC
'               0010 1xx0       POST INC
'               0011 0xx0       PRE  DEC
'               0011 1xx0       POST DEC
                
Using_3         and     opc,#%0000_1000 WZ, NR          ' If it's a post inc/dec push current value
        IF_NZ   call    #PrePush                        ' before altering the variable
                and     opc,#%0001_0000 WZ, NR
        IF_Z    add     lhs,#1
        IF_NZ   sub     lhs,#1

                shr     flags,#2 WC, WZ, NR             ' Recall size needed
  IF_Z_AND_NC   and     lhs,#$FF                        ' 00 BYTE
  IF_Z_AND_C    and     lhs,k_0000_FFFF                 ' 01 WORD

'               .---------------------------------------------------------------------------------------.
'               | Put result back - Note WC and WZ must have been set by SHR flags,#2 WC,WZ, NR         |                                                                      |
'               `---------------------------------------------------------------------------------------'

Using_4         call    #Put                            ' Put result to var
                call    #PrePush                        ' Do any push if we have not already done it
                
                jmp     #Fetch                          ' Next instruction

' .-----------------------------------------------------------------------------------------------------.
' |     Using Operator lookup table                                                                     |
' `-----------------------------------------------------------------------------------------------------'
'
' Each entry in this table can be one of four things ...
'
' 1) A native propeller instruction which operates on lhs, rhs or both.
'
' 2) A 'long OprLabel' which is the cog address of a handler within the cog.
'
' 3) A 'long @OvrLabel' which is the hub address of an overlay handler which will be loaded into the
'    cog and subsequently executed.
'
' 4) A 'long @OvrLabel | LMM' which is the hub address of an overlay handler which will be interpreted
'    Large Memory Model style.
'
' Overlays interpreted LMM-style cannot contain loops ( jumps to within themselves ), self-modifying
' code nor any local constants. Including any of those will lead to failure.

'                                                       '       ***        * ** 
Using_Table     long    @OvrRPTXXXJ                     ' ---- -000 <== ---0 0010 <== $02 RPTINCJ
                long    @OvrRPTXXXJ                     ' ---- -001 <== ---0 0110 <== $06 RPTADDJ
                long    @OvrFWDRAND                     ' ---- -010 <== ---0 1000 <== $08 FWDRAND
                long    @OvrREVRAND                     ' ---- -011 <== ---0 1100 <== $0C REVRAND

                long    @OvrPRESEXX | LMM               ' ---- -100 <== ---1 0000 <== $10 PRESEXB
                long    @OvrPRESEXX | LMM               ' ---- -101 <== ---1 0100 <== $14 PRESEXW
                long    @OvrPOSTXXX | LMM               ' ---- -110 <== ---1 1000 <== $18 POSTCLR
                long    @OvrPOSTXXX | LMM               ' ---- -111 <== ---1 1100 <== $1C POSTSET

                long    OprPREPOP                       ' ---- 1000 <== ---0 0000 <== $00 PREPOP

DAT
                
' *******************************************************************************************************
' *                                                                                                     *
' *     Operator Dispatch                                                                               *
' *                                                                                                     *
' *******************************************************************************************************
'
' All 'math operations' are done on parameters pushed to the stack and are a single byte long. The
' operands are pulled from the stack, processed then put back. There are a mix of unary and binary
' operators, including comparisons, and logical operators. The bytecode opcode mapping appears to be
' quite random and it's not easy to determine how the Parallax Spin Interpreter may be dealing with
' the opcode decoding and dispatching. They are handled in this VM by using a native instruction
' lookup table and special case handlers. To make decoding faster, the lookup table is currently hard
' coded to be at address $0E0-$0FF, giving a one-to-one mapping of opcode byte value to cog address.
'
' If we have a binary operator we put TOS into 'rhs' and NOS into 'lhs' so we can 'lhs := lhs op rhs'
' to make the handling esier and more logical. For unary ops we have TOS in both 'lhs' and 'rhs' to
' make life easier for the unary op handlers.
'
' If the operator is a logical operator we convert the popped values to booleans, that is a non-zero
' value becomes -1.
'
' To handle the opcode's intent we lookup a native opcode from $0E0-$0FF and execute it. For compare
' bytecodes we pre-compare 'lhs' and 'rhs' and preset the result to false so the native opcode
' looked-up can simple set a true based on its IF_XX conditional execution.
'
' For any bytecodes which are too complex to handle this way, we use a call  in the lookup table to 
' a more comprehensive handler.
'
'                .---.---.---.---.---.---.---.---.
'                | 1   1   1 | o   o   o   o   o |
'                `---^---^---^---^---^---^---^---'
'                 
'                E0  ROR          | E8  BIT_AND      | F0  LOG_AND      | F8  SQRT
'                E1  ROL          | E9  ABS          | F1  BIT_GET      | F9  LT
'                E2  SHR          | EA  BIT_OR       | F2  LOG_OR       | FA  GT
'                E3  SHL          | EB  BIT_XOR      | F3  BIT_SET      | FB  NE
'                E4  HIGHEST      | EC  ADD          | F4  MPY          | FC  EQ
'                E5  LOWEST       | ED  SUB          | F5  MPY_BIG      | FD  LE
'                E6  NEG          | EE  SAR          | F6  DIV          | FE  GE
'                E7  BIT_NOT      | EF  BIT_REV      | F7  MOD          | FF  LOG_NOT
'
' Operator types ...
'
'                        FFFFFFFF FFFFFFFF EEEEEEEE EEEEEEEE
'                        FEDCBA98 76543210 FEDCBA98 76543210
'                 
k_BinaryOps     long    %01111110_11110101_11111101_00111111
k_LogicalOps    long    %10000000_00000101_00000000_00000000
k_CompareOps    long    %01111110_00000000_00000000_00000000

Operator        call    #PopRhs                         ' Pop rhs ( and into lhs )

                movs    Opr_Src,opc                     ' Ready to get actual opcode

                mov     opcBit,#1                       ' Turn opcode into an opcode bit
                shl     opcBit,Opr_Src                       
                add     Opr_Src,#Operator_Table-$E0
                            
                and     opcBit,k_BinaryOps WZ, NR       ' Is it a binary op ?
        IF_NZ   call    #PopLhs                         ' Yes, Pop lhs

                jmpret  CommonReturn,#Opr_DoIt

Push            call    #PushLhs                        ' Put result to stack
                
                jmp     #Fetch                          ' Next instruction

' .-----------------------------------------------------------------------------------------------------.
' |     Operator Dispatcher                                                                             |
' `-----------------------------------------------------------------------------------------------------'

Opr_DoIt        and     opcBit,k_LogicalOps WZ, NR      ' Is it a logic op ?
        IF_NZ   cmp     lhs,#0 WZ                       ' Yes, so if lhs <> 0 ...
        IF_NZ   mov     lhs,k_FFFF_FFFF                 ' ... set lhs to -1

                and     opcBit,k_LogicalOps WZ, NR      ' Is it a logic op ?
        IF_NZ   cmp     rhs,#0 WZ                       ' Yes, so if rhs <> 0 ...
        IF_NZ   mov     rhs,k_FFFF_FFFF                 ' ... set rhs to -1

                and     opcBit,k_CompareOps WZ, NR      ' Is it a compare op ?
        IF_NZ   mov     opc,lhs
        IF_NZ   mov     lhs,#0
        IF_NZ   cmps    opc,rhs WZ, WC, NR              ' Yes, so compare signed

Opr_Src         mov     ovlPtr,0-0                      ' Get overlay to use

                andn    ovlPtr,k_0000_FFFF WZ, NR       ' If any of the MSW bits are set it is a PASM
        IF_NZ   mov     OVERLAY_END-1,ovlPtr            ' instruction placed at the end of the overlay
        IF_NZ   jmp     #OVERLAY_END-1                  ' and jumped to

                jmp     #Overlay_This                   ' Execute overlay

' .-----------------------------------------------------------------------------------------------------.
' |     Operator lookup table                                                                           |
' `-----------------------------------------------------------------------------------------------------'
'
' Each entry in this table can be one of four things ...
'
' 1) A native propeller instruction which operates on lhs, rhs or both.
'
' 2) A 'long OprLabel' which is the cog address of a handler within the cog.
'
' 3) A 'long @OvrLabel' which is the hub address of an overlay handler which will be loaded into the
'    cog and subsequently executed.
'
' 4) A 'long @OvrLabel | LMM' which is the hub address of an overlay handler which will be interpreted
'    Large Memory Model style.
'
' Overlays interpreted LMM-style cannot contain loops ( jumps to within themselves ), self-modifying
' code nor any local constants. Including any of those will lead to failure.

Operator_Table  ror     lhs,rhs                         ' $E0   ROR             ->
                rol     lhs,rhs                         ' $E1   ROL             <-
                shr     lhs,rhs                         ' $E2   SHR             >>
                shl     lhs,rhs                         ' $E3   SHL             <<
                mins    lhs,rhs                         ' $E4   MIN / HIGHEST   #>
                maxs    lhs,rhs                         ' $E5   MAX / LOWEST    <#
                neg     lhs,lhs                         ' $E6   NEG             -       Unary
                xor     lhs,k_FFFF_FFFF                 ' $E7   BIT_NOT         !       Unary                       
                and     lhs,rhs                         ' $E8   BIT_AND         &
                abs     lhs,lhs                         ' $E9   ABS             ||      Unary
                or      lhs,rhs                         ' $EA   BIT_OR          |
                xor     lhs,rhs                         ' $EB   BIT_XOR         ^
                add     lhs,rhs                         ' $EC   ADD             +
                sub     lhs,rhs                         ' $ED   SUB             -
                sar     lhs,rhs                         ' $EE   SAR             ~>
                long    @OvrBIT_REV        | LMM        ' $EF   BIT_REV         ><
                and     lhs,rhs                         ' $F0   LOG_AND         and
                long    @OvrBIT_GET                     ' $F1   BIT_GET         >|      Unary
                or      lhs,rhs                         ' $F2   LOG_OR          or
                long    OprBIT_SET                      ' $F3   BIT_SET         |<      Unary
                long    @OvrMPY                         ' $F4   MPY             *
                long    @OvrMPY                         ' $F5   MPY_BIG         **
                long    @OvrDIV                         ' $F6   DIV             /
                long    @OvrDIV                         ' $F7   MOD             //
                long    @OvrSQRT                        ' $F8   SQRT            ^^      Unary
        IF_B    mov     lhs,k_FFFF_FFFF                 ' $F9   LT              <
        IF_A    mov     lhs,k_FFFF_FFFF                 ' $FA   GT              >
        IF_NE   mov     lhs,k_FFFF_FFFF                 ' $FB   NE              <>
        IF_E    mov     lhs,k_FFFF_FFFF                 ' $FC   EQ              ==
        IF_BE   mov     lhs,k_FFFF_FFFF                 ' $FD   LE              =<
        IF_AE   mov     lhs,k_FFFF_FFFF                 ' $FE   GE              =>
                xor     lhs,k_FFFF_FFFF                 ' $FF   LOG_NOT         not     Unary

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Stack Interface                                                                                 *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |             Pop three items into acc, lhs and rhs                                                   |
' `-----------------------------------------------------------------------------------------------------'

PopAccThenRhsThenLhs     call   #PopLhs
                         mov    acc,lhs
                         call   #PopRhs
                         call   #PopLhs
PopAccThenRhsThenLhs_ret ret

' .-----------------------------------------------------------------------------------------------------.
' |             Pop from stack into lhs                                                                 |
' `-----------------------------------------------------------------------------------------------------'

PopLhs          rdlong  lhs,sp
                sub     sp,#4
PopLhs_Ret      ret

' .-----------------------------------------------------------------------------------------------------.
' |            Pop from stack into lhs and rhs                                                          |
' `-----------------------------------------------------------------------------------------------------'

PopRhs          call    #PopLhs
                mov     rhs,lhs
PopRhs_Ret      ret

' .-----------------------------------------------------------------------------------------------------.
' |             Push to stack                                                                           |
' `-----------------------------------------------------------------------------------------------------'

PushZero        mov     lhs,#0

PushLhs         add     sp,#4
                wrlong  lhs,sp
PushLhs_Ret

PushZero_Ret    ret

' .-----------------------------------------------------------------------------------------------------.
' |             Push sub-result to stack, but only if we've not already pusehd a sub-result             |
' `-----------------------------------------------------------------------------------------------------'

PrePush         and     flags,#2 WZ, NR                 ' Only do the Push if we have not done it

        IF_NZ   call    #PushLhs                        ' Do the push
                andn    flags,#2                        ' Clear the must push flag
                
PrePush_Ret     ret

' .-----------------------------------------------------------------------------------------------------.
' |             Sort lhs and rhs into ascending order ( ie, lhs =< rhs )                                |
' `-----------------------------------------------------------------------------------------------------'

'               Step                    What ends upin lhs    What end sup in rhs
'               xor     lhs,rhs         lhs^rhs               rhs
'               xor     rhs,lhs         lhs^rhs               rhs^lhs^rhs = lhs
'               xor     lhs,rhs         lhs^rhs^lhs = rhs     lhs           

SortLhsRhs      cmps    lhs,rhs WC, WZ, NR              ' Is lhs > rhs ?
        IF_A    xor     lhs,rhs                         ' Yes, swap lhs with rhs
        IF_A    xor     rhs,lhs
        IF_A    xor     lhs,rhs
SortLhsRhs_Ret  ret

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Hub Memory Interface                                                                            *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |             Read a variable from hub memory and then push it to the stack                           |
' `-----------------------------------------------------------------------------------------------------'

GetThenPush     call    #Get
                call    #PushLhs
GetThenPush_Ret ret

' .-----------------------------------------------------------------------------------------------------.
' |             Read a variable from hub memory into lhs ( or from SPR )                                |
' `-----------------------------------------------------------------------------------------------------'

Get
  IF_NZ_AND_C   jmp     #GetSpr                         ' 11 Special case for SPR access
  IF_Z_AND_NC   rdbyte  lhs,opr                         ' 00 BYTE
  IF_Z_AND_C    rdword  lhs,opr                         ' 01 WORD
  IF_NZ_AND_NC  rdlong  lhs,opr                         ' 10 LONG
Get_Ret         ret

GetSpr          movs    :Src,opr
                shl     bitRev,#1 WC, NR
:Src            mov     lhs,0-0                         ' Read the SPR

                cmp     opr,#$009 WZ, NR                ' Actually wanting CogId ?
        IF_Z    cogid   lhs                             ' Yes

                and     lhs,bitMsk                      ' Keep only the bits we want
                shr     lhs,bitShf                      ' Align the lsb
        IF_NC   rev     lhs,bitRev                      ' Reverse the bits as required
                jmp     #Get_Ret
                
' .-----------------------------------------------------------------------------------------------------.
' |             Pop a value from the stack then write it to a variable in hub memory                    |
' `-----------------------------------------------------------------------------------------------------'

PopThenPut      call    #PopLhs
                call    #Put
PopThenPut_Ret  ret

' .-----------------------------------------------------------------------------------------------------.
' |             Write lhs to a variable in hub memory ( or to SPR )                                     |
' `-----------------------------------------------------------------------------------------------------'

Put
  IF_NZ_AND_C   jmp     #PutSpr                         ' 11 Special case for SPR access
  IF_Z_AND_NC   wrbyte  lhs,opr                         ' 00 BYTE
  IF_Z_AND_C    wrword  lhs,opr                         ' 01 WORD
  IF_NZ_AND_NC  wrlong  lhs,opr                         ' 10 LONG
Put_Ret         ret

PutSpr          movd    :Dst,opr                        ' Point to the required SPR to update
                movs    :Src,opr
                mov     acc,lhs                         ' Don't mess with the sub expression result
                shl     bitRev,#1 WC, NR
        IF_NC   rev     acc,bitRev                      ' Reverse the bits as required
                shl     acc,bitShf                      ' Shift the bits to their destination position
                and     acc,bitMsk                      ' Only keep the bits we are affecting
:Src            mov     rhs,0-0                         ' Get the current SPR value
                andn    rhs,bitMsk                      ' Keep the bits we are not affecting
                or      rhs,acc                         ' Update the affected bits
:Dst            mov     0-0,rhs                         ' Put the result back
                jmp     #Put_Ret                        ' Return
                
DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Utility Code                                                                                    *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |           Get signed address branch                                                                 |
' `-----------------------------------------------------------------------------------------------------'
'
' OPTIMISE :    It would be possible to move the second rdbyte up the order ( not the 'add pc,#1' ) so it
'               pre-empts the second byte read without a long delay between hub accesses.

GetAddress      rdbyte  adr,pc
                add     pc,#1

                mov     acc,#25
                
                and     adr,#%10000_0000 WZ, NR
                
        IF_NZ   rdbyte  rhs,pc
        IF_NZ   add     pc,#1
        IF_NZ   shl     adr,#8
        IF_NZ   or      adr,rhs

        IF_NZ   mov     acc,#17

                shl     adr,acc
                sar     adr,acc

GetAddress_Ret  ret

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Favoured Command Opcode Handlers ( Non-Overlay )                                                *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |     $00     FRAME   0       a := MyFunc                                                             |
' |     $01     FRAME   1       MySub                                                                   |
' |     $02     FRAME   2       a := \MyFunc                                                            |
' |     $03     FRAME   3       \MySub                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
' On Entry ...
'
'               .------.------.
'     BP -----> | XXXX | XXXX |
'               |------^------|
'               | 0000   0000 |
'               |-------------|
'               | local var 1 |
'               |-------------|
'     SP -----> | local var N |
'               `-------------'
'
' On Exit ...
'
'               .------.------.                 
'     BP -----> | XXXX | XXXX |               
'               |------^------|               
'               | 0000   0000 |               <-------.
'               |-------------|                       | 
'               | local var 1 |                       | 
'               |-------------|                       | 
'               | local var N |                       | 
'               `-------------|------.------.         |
'                             |  op  | vp f |         |   Adds this
'                             |------+------|         |
'     Next BP --------------> | pnbp | bp+4 |---------'   Adds this, Saves prev Next BP, Sets Next BP
'                             |------^------|
'     SP -------------------> | 0000   0000 |             Adds this ( result value )
'                             `-------------'

OpcFRAME        add     sp,#4                           ' Point to "op | vp f"
                wrword  op,sp                           ' Set "op"
        
                add     sp,#2                           ' Point to "vp f"
                and     opc,#3                          ' Mix VP with framing info
                or      opc,vp
                wrword  opc,sp                          ' Set Framing Info

                add     sp,#2                           ' Point to "pnbp | bp+4"
                mov     rhs,nextBp                      ' Remember previous BP value  
                mov     nextBp,sp                       ' Mark next BP for this frame
                wrlong  rhs,nextBp                      ' Put previous next BP in frame
                
                add     sp,#2                           ' Point to "bp+4"

                mov     acc,bp                          ' Set 'bp+4'
                add     acc,#4
                wrword  acc,sp
                
                add     sp,#2                           ' Zero the result value
                wrword  k_0000_0000,sp

                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |     $05     CALL                                                                                    |
' `-----------------------------------------------------------------------------------------------------'
'
' On Entry ...
'
'               .------.------.                 
'     BP -----> | XXXX | XXXX |               
'               |------^------|                       
'               | 0000   0000 |               <-------.
'               |-------------|                       |     
'               | local var 1 |                       |     
'               |-------------|                       |     
'               | local var N |                       | 
'               `-------------|------.------.         |
'                             |  op  | vp f |         |
'                             |------+------|         |
'     Next BP --------------> | pnbp | bp+4 |---------'
'                             |------^------|
'                             | 0000   0000 |
'                             `-------------|-------------.
'                                           | parameter 1 |
'                                           |-------------|
'     SP ---------------------------------> | parameter N |
'                                           `-------------'
'
' On Exit ...
'
'               .------.------.                 
'               | XXXX | XXXX |                   
'               |------^------|                       
'               | 0000   0000 |               <-------.
'               |-------------|                       |     
'               | local var 1 |                       |     
'               |-------------|                       |     
'               | local var N |                       | 
'               `-------------|------.------.         |
'                             |  op  | vp f |         |
'                             |------+------|         |                   Restores previous Next BP
'     BP -------------------> |  pc  | bp+4 |---------'                   Sets return pc, Sets BP
'                             |------^------|
'                             | 0000   0000 |                              
'                             `-------------|-------------.
'                                           | parameter 1 |
'                                           |-------------|
'                                           | parameter N |
'                                           `-------------|-------------.
'                                                         | local var 1 | Adds this
'                                                         |-------------|
'     SP -----------------------------------------------> | local var N | Adds this
'                                                         `-------------'

OpcCALL         rdbyte  acc,pc                          ' Get the subroutine call index
                add     pc,#1

                shl     acc,#2                          ' Convert to a long offset
                add     acc,op                          ' Adjust for object base
                
                rdword  lhs,acc                         ' Read the destination pc ( word )
                add     lhs,op                          ' Adjust for object base

                add     acc,#2                          ' Read the sp increment value ( local vars )
                rdword  rhs,acc

'               lhs   = pc of subroutine
'               rhs   = local var stack increment

                mov     bp,nextBp                       ' Adjust BP to new frame
                rdword  nextBp,bp                       ' Restore previous Next BP
                
                wrword  pc,bp                           ' Put return address on stack

                add     sp,rhs                          ' Add local vars

                mov     pc,lhs                          ' Continue from called pc

                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |     $08     LOOPJPF                                                                                 |
' |     $09     LOOPRPT                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

OpcLOOPXXX      call    #GetAddress

                call    #PopLhs

                sub     lhs,#0 WZ, NR
                shr     opc,#1 WC, NR
                 
    IF_C        sub     lhs,#1 WZ

    IF_NZ       call    #PushLhs
    IF_C_NE_Z   add     pc,adr

                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |     $30     ABORT                                                                                   |
' |     $31     ABOVAL                                                                                  |
' |     $32     RETURN                                                                                  |
' |     $33     RETVAL                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
' On Entry ...
'                                            
'               .------.------.                 
'               | XXXX | XXXX |               
'               |------^------|                  
'               | 0000   0000 |               <-------.
'               |-------------|                       |     
'               | local var 1 |                       |     
'               |-------------|                       |     
'               | local var N |                       |
'               `-------------|------.------.         |
'                             |  op  | vp f |         |
'                             |------+------|         |
'     BP -------------------> |  pc  | bp+4 |---------'
'                             |------^------|
'                             | result valu |
'                             `-------------|-------------.
'                                           | parameter 1 |
'                                           |-------------|
'                                           | parameter N |
'                                           `-------------|-------------.
'                                                         | local var 1 |
'                                                         |-------------|
'                                                     .-> | local var N |
'     SP --------------------------------------------<    `-------------|-------------.
'                                                     `->               | return valu |
'                                                                       `-------------'
'                                            
' After popping return value from stack or taking from 'result' variable we can ignore what
' is on the stack and re-build where we were from just the base pointer ...
'
'               .------.------.                 
'               | XXXX | XXXX |               
'               |------^------|                  
'               | 0000   0000 |               <-------.
'               |-------------|                       |     
'               | local var 1 |                       |     
'               |-------------|                       |     
'               | local var N |                       |
'               `-------------|------.------.         |
'                             |  op  | vp f |         |
'                             |------+------|         |
'     BP -------------------> |  pc  | bp+4 |---------'
'                             `------^------'
' On exit ...
'
'               .------.------.       
'     BP -----> | XXXX | XXXX |       
'               |------^------|       
'               | 0000   0000 |       
'               |-------------|       
'               | local var 1 |       
'               |-------------|       
'           .-> | local var N |       
'     SP --<    `-------------|-------------.
'           `->               | return valu |
'                             `-------------'

OpcRETURN       shr     opc,#1 WC, NR           ' Is the return value on stack ?
        IF_C    rdlong  lhs,sp                  ' Yes, read from stack
        IF_NC   add     bp,#4                   ' No, it's in result variable
        IF_NC   rdlong  lhs,bp                  ' Read result
        IF_NC   sub     bp,#4

:Loop           rdword  pc,bp                   ' Set the return PC

                sub     bp,#2                   ' Restore VP / framing info
                rdword  vp,bp

                sub     bp,#2                   ' Restore OP
                rdword  op,bp

                mov     sp,bp                   ' Restore SP
                sub     sp,#4
                
                add     bp,#6                   ' Restore BP
                rdword  bp,bp

                cmp     bp,k_0000_FFFF WZ, NR   ' Exited main program ?
        IF_Z    cogid   lhs                     ' Yes, Get our Cog Id ...
        IF_Z    cogstop lhs                     ' ... And stop ourselves running

                sub     bp,#4                   ' Adjust BP to where we need it

                and     opc,#%0000_0010 WZ, NR  ' Is it an abort opcode ?
        IF_NZ   jmp     #:Done                  ' No - We simple return

                and     vp,#%0000_00010 WZ, NR  ' Was it an abort trap call ?
        IF_Z    jmp     #:Loop                  ' No - Keep popping through returns
                          
:Done           shr     vp,#1 WC, NR            ' Are we discarding return value ?
                andn    vp,#3

        IF_NC   call    #PushLhs                ' No, push return value to stack
                            
                jmp     #Fetch

' .-----------------------------------------------------------------------------------------------------.
' |     $34     PUSH    #-1                                                                             |
' |     $35     PUSH    #0                                                                              |
' |     $36     PUSH    #1                                                                              |
' `-----------------------------------------------------------------------------------------------------'

OpcPUSH_K       mov     lhs,opc
                sub     lhs,#$35
                jmp     #Push

' .-----------------------------------------------------------------------------------------------------.
' |     $37     PUSH    #kp                                                                             |
' `-----------------------------------------------------------------------------------------------------'

OpcPUSH_Kp      rdbyte  rhs,pc
                add     pc,#1
                mov     lhs,#2
                shl     lhs,rhs

                and     rhs,#%0010_0000 WZ, NR
        IF_NZ   sub     lhs,#1

                and     rhs,#%0100_0000 WZ, NR
        IF_NZ   xor     lhs,k_FFFF_FFFF

                jmp     #Push

' .-----------------------------------------------------------------------------------------------------.
' |     $38     PUSH    #k1                                                                             |
' |     $39     PUSH    #k2                                                                             |
' |     $3A     PUSH    #k3                                                                             |
' |     $3B     PUSH    #k4                                                                             |
' `-----------------------------------------------------------------------------------------------------'

OpcPUSH_Kn      mov     lhs,#0
                sub     opc,#$37

:Loop           rdbyte  rhs,pc
                add     pc,#1
                shl     lhs,#8
                or      lhs,rhs
                djnz    opc,#:Loop

                jmp     #Push

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Favoured Using Operator Handlers ( Non-Overlay )                                                *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |     $00     PREPOP  a := ( b := ( 1 ) )                                                             |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; a := ( b := 1 )
'               
'               0018    36                                PUSH     1
'               0019    46 80                             USING    VAR+4.LONG PUSH PREPOP
'               001B    41                                POP      VAR+0.LONG
               
OprPREPOP       call    #PopThenPut
                
                jmp     #CommonReturn

' .-----------------------------------------------------------------------------------------------------.
' |     $18     POSTCLR lhs~                                                                            |
' |     $1C     POSTSET lhs~~                                                                           |
' `-----------------------------------------------------------------------------------------------------'

OvrPOSTXXX      call    #PrePush
                and     opc,#%0000_0100 WZ, NR
                muxnz   lhs,k_FFFF_FFFF

                jmp     #CommonReturn

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Favoured Operator Handlers ( Non-Overlay )                                                      *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |     $F3     BIT_SET lhs := |< rhs, lhs := 1 << rhs                                                  |
' `-----------------------------------------------------------------------------------------------------'

OprBIT_SET      mov     lhs,#1                          ' 1 << rhs
                shl     lhs,rhs

                jmp     #CommonReturn
                
DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Pre-defined Constants                                                                           *
' *                                                                                                     *
' *******************************************************************************************************

k_0000_0000     long    $0000_0000                      ' 0 / False
k_0000_0200     long    $0000_0200
k_0000_FFFF     long    $0000_FFFF
k_FFFF_FFFF     long    $FFFF_FFFF                      ' -1 / True

' *******************************************************************************************************
' *                                                                                                     *
' *     Run-Time variables                                                                              *
' *                                                                                                     *
' *******************************************************************************************************

pc              long    0                               ' Program counter
op              long    0                               ' Object base pointer ( subs table )
vp              long    0                               ' Variable base pointer
bp              long    0                               ' Stack base pointer ( locals )
sp              long    0                               ' Stack pointer ( points to TOS )
nextBp          long    0                               ' Next base pointer
cogBp           long    $0010                           ' Cog Base Pointer

ovlPtr          long    @SpinVm                         ' Overlay loading pointer
ovlCnt          long    0                               ' Overlay counter

opc             long    0                               ' Opcode
opr             long    0                               ' Operand / Operand's address
adr             long    0                               ' Address offset
opcBit          long    0                               ' 1 << ( opc & $1F ), used in Operator/Using

lhs             long    0
rhs             long    0
acc             long    0

flags           long    0

'               These are used for DIRA[7..0] etc operations ...

bitMsk          long    0                               ' Bit field mask
bitShf          long    0                               ' Bit shift count
bitRev          long    0                               ' Bit reverse count

' .-----------------------------------------------------------------------------------------------------.
' |             Local registers allocated for overlays to use                                           |
' `-----------------------------------------------------------------------------------------------------'

REGISTERS       long    0[ REGISTERS_SIZE-1 ]

LAST_REGISTER   long    0[ ( ($1F0/4)-$ ) #> 1 ]

                fit     $1F0

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Overlays                                                                                        *
' *                                                                                                     *
' *******************************************************************************************************
'
' All overlays should start with 'org OVERLAY_START' and end with 'fit OVERLAY_SIZE' to ensure they fit
' in the overlay area. The overlay loader will always load OVERLAY_SIZE longs unless it encounters a
' 'long 0' after which it will stop loading longs. Putting a 'long 0' after the last instruction of the
' overlay will therefore speed up loading and overall execution time. The 'long 0' will be replaced with
' a 'jmp #CommonReturn' which would normally be at the end of the overlay so it does not waste any hub
' memory.
'
' The OVERLAY_SIZE will ideally be adjusted so overlays can 'fall through' to the CommonReturn point,
' and the variable size means that the 'long 0' might not fit for some OVERLAY_SIZE settings but would
' be required for larger OVERLAY_SIZE values. The ideal 'long 0' to use is as follows ...
'
'               long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
'
' This will generate a single long when required but none when it is not. All overlays will therefore
' usually have the following format -
'
'               org     OVERLAY_START                           Start the overlay                             
'
'       Label   code                                            The overlay's code
'               :
'               code
'
'       :org    org     REGISTERS                               Define local registers
'       :var1   res     1
'       :var2   res     1
'               fit     LAST_REGISTER+1                         Check not too many local registers
'               org     :org
'
'               long    0[ 0 #> (OVERLAY_END-$) <# 1 ]          End with a 'long 0' if necessary
'               fit     OVERLAY_END                             Check overlay isn't too long
                
DAT
                
' *******************************************************************************************************
' *                                                                                                     *
' *     Command Opcode Handler Overlays                                                                 *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |     $04     GOTO                                                                                    |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcGOTO         call    #GetAddress
                add     pc,adr
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $06     CALLOBJ                                                                                 |
' |     $07     CALLOBJ[]                                                                               |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvcCALLOBJ      rdbyte  acc,pc
                add     pc,#1

                cmp     opc,#$07 WZ,NR
        IF_Z    call    #PopLhs
        IF_Z    add     acc,lhs

                shl     acc,#2
                add     acc,op
                rdword  lhs,acc
                add     lhs,op

                add     acc,#2
                rdword  acc,acc

                mov     op,lhs
                add     vp,acc
                
                jmp     #OpcCall

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $0A     JPF                                                                                     |
' |     $0B     JPT                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvcJPX          call    #GetAddress

                call    #PopLhs

                sub     lhs,#0 WZ, NR
                shr     opc,#1 WC, NR
   IF_C_NE_Z    add     pc,adr

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $0C     CASEJMP                                                                               |
' `-----------------------------------------------------------------------------------------------------'

' Case compares the most recently pushed value (TOS) with the value previously pushed (NOS). Note that
' it does not pull the NOS from stack; that is left for CASEJMP to deal with.
'
'               ====                                    ; case a
'               ====                                    ;   0 : a := 1
'               ====                                    ;   1 : a := 2
'
'               0018    38 19                             PUSH     25
'               001A    40                                PUSH     VAR+0.LONG
'               001B    35                                PUSH     0
'               001C    0D 04                             CASE     +4
'               001E    36                                PUSH     1
'               001F    0D 04                             CASE     +4
'               0021    0C                                CASEJMP
'               0022    36                      J4        PUSH     1
'               0023    41                                POP      VAR+0.LONG
'               0024    0C                                CASEJMP
'               0025    37 00                   J5        PUSH     2
'               0027    41                                POP      VAR+0.LONG
'               0028    0C                                CASEJMP
'               0029    32 J6                             RETURN   

                org     OVERLAY_START

OvcCASEJMP      sub     sp,#4                           ' Pop the <expr> and discard
                call    #PopLhs                         ' Pop the branch offset
                add     lhs,op
                mov     pc,lhs

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $0D     CASE    Case single value                                                               |
' |     $0E     CASERNG Case range                                                                      |
' `-----------------------------------------------------------------------------------------------------'
'
' Case compares the most recently pushed value (TOS) with the value previously pushed (NOS). Note that
' it does not pull the NOS from stack; that is left for CASEJMP to deal with.

'               ====                                    ; case a
'               ====                                    ;   1    : a := 1
'               ====                                    ;   2    : a := 2
'               ====                                    ;   3..5 : a := 325
'               ====                                    ;   8..6 : a := 628
'
'               0018    38 30                             PUSH     $30
'               001A    40                                PUSH     VAR+0.LONG
'               001B    36                                PUSH     1
'               001C    0D 11                             CASE     +17
'               001E    37 00                             PUSH     2
'               0020    0D 10                             CASE     +16
'               0022    37 21                             PUSH     3
'               0024    38 05                             PUSH     5
'               0026    0E 0E                             CASERNG  +14
'               0028    37 02                             PUSH     8
'               002A    38 06                             PUSH     6
'               002C    0E 0D                             CASERNG  +13
'               002E    0C                                CASEJMP
'               002F    36                      J4        PUSH     1
'               0030    41                                POP      VAR+0.LONG
'               0031    0C                                CASEJMP
'               0032    37 00                   J5        PUSH     2
'               0034    41                                POP      VAR+0.LONG
'               0035    0C                                CASEJMP
'               0036    39 01 45                J6        PUSH     325
'               0039    41                                POP      VAR+0.LONG
'               003A    0C                                CASEJMP
'               003B    39 02 70                J7        PUSH     624
'               003E    41                                POP      VAR+0.LONG
'               003F    0C                                CASEJMP
'               0040    32                      J8        RETURN
 
                org     OVERLAY_START

OvcCASE         call    #GetAddress                     ' Read the address we could branch to
                call    #PopRhs                         ' Get single value to lhs and rhs

                shr     opc,#1 WC, NR                   ' Is it a range ?
        IF_NC   call    #PopLhs                         ' Yes, Pop lhs
        IF_NC   call    #SortLhsRhs                     ' Yes, Sort lhs and rhs into asceding order

                rdlong  acc,sp                          ' Get <expr> to match with ( don't pop )

                cmps    acc,lhs WZ, WC, NR              ' Is it less than lhs ?
        IF_B    jmp     #CommonReturn                   ' Yes, not in range

                cmps    acc,rhs WZ, WC, NR              ' Is it greater than rhs ?
        IF_A    jmp     #CommonReturn                   ' Yes, not in range
                
                add     pc,adr                          ' In range, jump to the address

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $0F     LOOKEND                                                                                 |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; a := lookup( a : "ABC" .. "F" )
'                
'               0018    36                                PUSH     1            PUSH 0 for lookupz
'               0019    38 18                             PUSH     $18
'               001B    40                                PUSH     VAR+0.LONG
'               001C    38 41                             PUSH     65
'               001E    10                                LOOKUP
'               001F    38 42                             PUSH     66
'               0021    10                                LOOKUP
'               0022    38 43                             PUSH     67
'               0024    38 46                             PUSH     70
'               0026    12                                LOOKUPR
'               0027    0F                                LOOKEND
'               0028    41                                POP      VAR+0.LONG

                org     OVERLAY_START

OvcLOOKEND      sub     sp,#4
                rdlong  pc,sp
                sub     sp,#4
                wrlong  k_0000_0000,sp
                add     pc,op

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $10     LOOKUP  Lookup single value                                                             |
' |     $12     LOOKUPR Lookup range                                                                    |
' `-----------------------------------------------------------------------------------------------------'

'               ====                                    ; a := lookup( a : "ABC" .. "F" )
'                
'               0018    36                                PUSH     1            PUSH 0 for lookupz
'               0019    38 18                             PUSH     $18
'               001B    40                                PUSH     VAR+0.LONG
'               001C    38 41                             PUSH     65
'               001E    10                                LOOKUP
'               001F    38 42                             PUSH     66
'               0021    10                                LOOKUP
'               0022    38 43                             PUSH     67
'               0024    38 46                             PUSH     70
'               0026    12                                LOOKUPR
'               0027    0F                                LOOKEND
'               0028    41                                POP      VAR+0.LONG

                org     OVERLAY_START

OvcLOOKUPX      call    #PopRhs                         ' Get single or range of values to match
                and     opc,#%0000_0010 WZ, NR
        IF_NZ   call    #PopLhs

                rdlong  acc,sp                          ' Read the variable's current value

                sub     sp,#8                           ' Get the base offset
                rdlong  opcBit,sp 
                
:Loop           cmp     acc,opcBit WZ,NR                ' Update the index

        IF_Z    wrlong  lhs,sp
        IF_Z    add     sp,#4
        IF_Z    rdlong  pc,sp
        IF_Z    sub     sp,#4
        IF_Z    add     pc,op
        IF_Z    jmp     #Fetch

                sub     acc,#1

                cmp     lhs,rhs WC, WZ, NR
        IF_B    add     lhs,#1
        IF_A    sub     lhs,#1
        IF_NZ   jmp     #:Loop
        
                add     sp,#8
                wrlong  acc,sp
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $11     LOOKDN  Lookdown single value                                                           |
' |     $13     LOOKDNR ookdown range                                                                   |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcLOOKDNX      call    #PopRhs                         ' Get single or range of values to match
                and     opc,#%0000_0010 WZ, NR
        IF_NZ   call    #PopLhs

                rdlong  acc,sp                          ' Read the variable's current value

                sub     sp,#8
                rdlong  opcBit,sp

:Loop           cmp     acc,lhs WZ, NR                  ' Found a match ?

         IF_Z   add     sp,#4
         IF_Z   rdlong  pc,sp
         IF_Z   sub     sp,#4
         IF_Z   add     pc,op
         IF_Z   jmp     #Fetch

                add     opcBit,#1
                wrlong  opcBit,sp

                cmp     lhs,rhs WC, WZ, NR
        IF_B    add     lhs,#1
        IF_A    sub     lhs,#1
        IF_NZ   jmp     #:Loop
                
                add     sp,#8

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $14     QUIT                                                                                    |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; repeat
'               ====                                    ;   case a
'               ====                                    ;   0 : quit
'                
'               0018         38 15                        PUSH     21
'               001A         40                           PUSH     VAR+0.LONG
'               001B         35                           PUSH     0
'               001C         0D 01                        CASE     +1
'               001E         0C                           CASEJMP
'               001F         37 02           J4           PUSH     8            ' How much to pop
'               0021         14                           QUIT
'               0022         04 03                        GOTO     +3           ' Abort
'               0024         0C                           CASEJMP
'               0025         04 71           J5           GOTO     -15

                org     OVERLAY_START

OvcQUIT         rdlong  acc,sp                          ' Read the stack adjust value
                sub     sp,acc                          ' And Adjust

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $15     MARK                                                                                    |
' `-----------------------------------------------------------------------------------------------------'
'
' NOTE : It might be "3F 8F 37" then "61" ( "POP LOC+0.WORD" )
'
'               ====                                    ; cogInit(1,MySub,@Stack)
'               
'               001C         37 00                        PUSH     2
'               001E         43                           PUSH     #VAR+0.LONG
'               001F         15                           MARK
'               0020         36                           PUSH     1
'               0021         3F 8F                        ?
'               0023         37 61                        PUSH     $FFFFFFFC
'               0025         D1                           POP      MEM[][].LONG
'               0026         2C                           COGISUB
'
'               ====                                    ; cogInit(1,MySub($A,$B),@Stack)
'                
'               001C         38 0A                        PUSH     10
'               001E         38 0B                        PUSH     11
'               0020         39 02 02                     PUSH     514
'               0023         43                           PUSH     #VAR+0.LONG
'               0024         15                           MARK
'               0025         36                           PUSH     1
'               0026         3F 8F                        ?
'               0028         37 61                        PUSH     $FFFFFFFC
'               002A         D1                           POP      MEM[][].LONG
'               002B         2C                           COGISUB

                org     OVERLAY_START

OvcMARK
jmp #NotImplemented

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $16     STRSIZE                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcSTRSIZE      call    #PopRhs
                mov     lhs,k_FFFF_FFFF

:Loop           rdbyte  opr,rhs
                add     lhs,#1
                add     rhs,#1
                tjnz    opr,#:Loop
                jmp     #Push
              
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $17     STRCOMP                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcSTRCOMP      call    #PopRhs
                call    #PopLhs

:Loop           rdbyte  :lhsChar,lhs
                add     lhs,#1
                rdbyte  :rhsChar,rhs
                add     rhs,#1

                cmp     :lhsChar,:rhsChar WZ, NR
        IF_Z    tjnz    :lhsChar,#:Loop

                muxz    lhs,k_FFFF_FFFF
                jmp     #Push

:org            org     REGISTERS
:lhsChar        res     1
:rhsChar        res     1
                fit     LAST_REGISTER+1
                org     :org
                                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $18     BYTEFIL                                                                                 |
' |     $19     WORDFIL                                                                                 |
' |     $1A     LONGFIL                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcXXXXFIL      call    #PopAccThenRhsThenLhs           ' acc=Count / rhs=value / lhs=Address
        
                tjz     acc,#CommonReturn               ' Nothing to do

                mov     opr,lhs                         ' Address must be in opr
                mov     lhs,rhs                         ' Value to write must be in lhs

                and     opc,#%0000_0011                 ' Set Z and C to size so we can call the 
                shr     opc,#1 WC, WZ, NR               ' Using_Put routine to do the write

                mov     rhs,#1                          ' Determine increment
                shl     rhs,opc
                
'               lhs   = value to write
'               opr   = address
'               acc   = count
'               rhs   = opr increment
     
:Loop           call    #Put
                add     opr,rhs
                djnz    acc,#:Loop
                   
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $1B     WAITPEQ                                                                                 |
' |     $1F     WAITPNE                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvcWAITPXX      call    #PopAccThenRhsThenLhs           ' acc=Port / rhs=mask / lhs=state
                shr     acc,#1 WC, NR
                and     opc,#%000_0100 WZ, NR
        IF_Z    waitpeq lhs,rhs
        IF_NZ   waitpne lhs,rhs
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END
                
' .-----------------------------------------------------------------------------------------------------.
' |     $1C     BYTEMOV                                                                                 |
' |     $1D     WORDMOV                                                                                 |
' |     $1E     LONGMOV                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcXXXXMOV      call    #PopAccThenRhsThenLhs           ' acc=Count / rhs=From / lhs=To
        
                tjz     acc,#CommonReturn               ' Nothing to do

                mov     :to,lhs                         
                mov     :from,rhs

                and     opc,#%0000_0011                 ' Determine step increment 
                mov     rhs,#1
                shl     rhs,opc

                cmp     :to,:from WC, WZ, NR            ' Is to > from ?
        IF_A    mov     :add,acc                        ' Determine offset to reverse copy ...
        IF_A    sub     :add,#1                         ' ... add := (count-1) * (opc&3)
        IF_A    shl     :add,opc
        IF_A    add     :to,:add
        IF_A    add     :from,:add
        IF_A    neg     rhs,rhs                         ' Negative increment

                shr     opc,#1 WC, WZ, NR               ' Ensure Get/Put routine for correct size
                
'               :to   = to addres
'               :from = from address
'               acc   = count
'               rhs   = opr increment / decrement
     
:Loop           mov     opr,:from
                call    #Get
                mov     opr,:to
                call    #Put
                add     :to,rhs
                add     :from,rhs
                djnz    acc,#:Loop

:org            org     REGISTERS
:to             res     1
:from           res     1
:add            res     1
                fit     LAST_REGISTER+1
                org     :org
                                    
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $20     CLKSET                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; clkset( %01101100, 4_000_000 )
'               
'               001C         38 6C                        PUSH   108
'               001E         3A 3D 09 00                  PUSH   4000000
'               0022         20                           CLKSET
'
'               ====                                    ; reboot

'               0023         37 06                        PUSH   $80
'               0025         35                           PUSH   0
'               0026         20                           CLKSET

                org     OVERLAY_START

OvcCLKSET       call    #PopRhs                         ' Get the frequency
                wrlong  rhs,#0                          ' update the global CLKFREQ variable

                call    #PopLhs                         ' Get the actual clock setting
                clkset  lhs

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $21     COGSTOP CogStop(id)                                                                     |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcCOGSTOP      call    #PopLhs

                cogstop lhs

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $22     LRETSUB LockRet                                                                         |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcLRETSUB      call    #PopLhs                         ' Get lock number to return

                lockret lhs

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $23     WAITCNT                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvcWAITCNT      call    #PopLhs

                waitcnt lhs,#0

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $24     PUSH    SPR[]                                                                           |
' |     $25     POP     SPR[]                                                                           |
' |     $26     USING   SPR[]                                                                           |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; a := SPR[2]   ' a := INA
'               ====                                    ; SPR[4] := a   ' OUTA := a
'               ====                                    ; SPR[4]++      ' OUTA++
'               
'               0018    37 00                   S3        PUSH     2
'               001A    24                                PUSH     SPR[]
'               001B    41                                POP      VAR+0.LONG
'
'               001C    40                                PUSH     VAR+0.LONG
'               001D    37 01                             PUSH     4
'               001F    25                                POP      SPR[]
'
'               0020    37 01                             PUSH     4
'               0022    26 2E                             USING    SPR[] POSTINC
'
'                       **
'               %0010 0100      Push  SPR[]
'               %0010 0101      Pop   SPR[]
'               %0010 0110      Using SPR[]
          
                org     OVERLAY_START
                
OvcXXX_SPR      call    #PopLhs
                mov     opr,lhs
              
                or      opc,#%0110_0000                 ' Force the Get/Put to use mov's not rdxxxx
                mov     bitMsk,k_FFFF_FFFF              ' Full width word usage
                mov     bitShf,#0
                mov     bitRev,#0
                                                
                and     opr,#$00F
                or      opr,#$1F0
                jmp     #Access_Spr_Here

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END
                
' .-----------------------------------------------------------------------------------------------------.
' |     $27     WAITVID                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcWAITVID      call    #PopRhs
                call    #PopLhs

                waitvid lhs,rhs
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $28     COGIFUN                                                                                 |
' |     $2C     COGISUB                                                                                 |
' `-----------------------------------------------------------------------------------------------------'
'
' NOTE : It might be "3F 8F 37" then "61" ( "POP LOC+0.WORD" )
'
'               ====                                    ; cogInit(1,@MyPasm,0)
'               ====                                    ; cognew(@MyPasm,0)
'               ====                                    ; a := cognew(@MyPasm,0)
'               
'               0020         36                           PUSH     1
'               0021         C7 08                        PUSH     #OBJ+8.LONG
'               0023         35                           PUSH     0
'               0024         2C                           COGISUB
'
'               0025         34                           PUSH     -1
'               0026         C7 08                        PUSH     #OBJ+8.LONG
'               0028         35                           PUSH     0
'               0029         2C                           COGISUB
'
'               002A         34                           PUSH     -1
'               002B         C7 08                        PUSH     #OBJ+8.LONG
'               002D         35                           PUSH     0
'               002E         28                           COGIFUN
'               002F         C9 81 90                     POP      VAR+400.LONG

'               ====                                    ; cogInit(1,MySub,@Stack)
'               ====                                    ; cognew(MySub,@stack)
'               ====                                    ; a := cognew(MySub,@stack)
'               
'               001C         37 00                        PUSH     2
'               001E         43                           PUSH     #VAR+0.LONG
'               001F         15                           MARK
'               0020         36                           PUSH     1
'               0021         3F 8F                        ?
'               0023         37 61                        PUSH     $FFFFFFFC
'               0025         D1                           POP      MEM[][].LONG
'               0026         2C                           COGISUB
'
'               0027         37 00                        PUSH     2
'               0029         43                           PUSH     #VAR+0.LONG
'               002A         15                           MARK
'               002B         2C                           COGISUB
'
'               002C         37 00                        PUSH     2
'               002E         43                           PUSH     #VAR+0.LONG
'               002F         15                           MARK
'               0030         28                           COGIFUN
'               0031         C9 81 90                     POP      VAR+400.LONG 
'
'               ====                                    ; cogInit(1,MySub($A,$B),@Stack)
'               ====                                    ; cognew(MySub($A,$B),@stack)
'               ====                                    ; a := cognew(MySub($A,$B),@stack)
'                
'               001C         38 0A                        PUSH     10
'               001E         38 0B                        PUSH     11
'               0020         39 02 02                     PUSH     514
'               0023         43                           PUSH     #VAR+0.LONG
'               0024         15                           MARK
'               0025         36                           PUSH     1
'               0026         3F 8F                        ?
'               0028         37 61                        PUSH     $FFFFFFFC
'               002A         D1                           POP      MEM[][].LONG
'               002B         2C                           COGISUB
'
'               002C         38 0A                        PUSH     10
'               002E         38 0B                        PUSH     11
'               0030         39 02 02                     PUSH     514
'               0033         43                           PUSH     #VAR+0.LONG
'               0034         15                           MARK
'               0035         2C                           COGISUB
'
'               0036         38 0A                        PUSH     10
'               0038         38 0B                        PUSH     11
'               003A         39 02 02                     PUSH     514
'               003D         43                           PUSH     #VAR+0.LONG
'               003E         15                           MARK
'               003F         28                           COGIFUN
'               0040         C9 81 90                     POP      VAR+400.LONG

                org     OVERLAY_START

OvcCOGIXXX      call    #PopAccThenRhsThenLhs           ' acc=ParamPtr / rhs=Address / lhs=CogId (-1=Self)


                cmp     lhs,k_FFFF_FFFF WZ, NR          ' Is CogId = -1, Self ?
        IF_Z    cogid   lhs                             ' Yes, get own Cog Id   

jmp #NotImplemented

                and     opc,#%0000_0100 WZ,NR           ' Is it a function ?
        IF_Z    jmp     #Push                           ' Yes
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $29     LNEWFUN LockNew - Returns 0..7, -1 if none available                                    |
' |     $2D     LNEWSUB                                                                                 |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcLNEWXXX      locknew lhs WC, WR                      ' lhs = 0..7, C = None available
        IF_C    mov     lhs,k_FFFF_FFFF
        
                and     opc,#%0000_0100 WZ,NR           ' Is it a function ?
        IF_Z    jmp     #Push                           ' Yes

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $2A     LSETFUN LockSet(id) - Returns true / false previous setting                             |
' |     $2E     LSETSUB                                                                                 |
' |     $2B     LCLRFUN LockClr(id) - Returns true / false previous setting                             |
' |     $2F     LCLRSUB                                                                                 |
' `-----------------------------------------------------------------------------------------------------'
'
'               0010 1010 = ---- -010 | ---- -1-- = ---- -110  LockSet
'               0010 1110 = ---- -110 | ---- -1-- = ---- -110  LockSet
'               0010 1011 = ---- -011 | ---- -1-- = ---- -111  LockClr
'               0010 1111 = ---- -111 | ---- -1-- = ---- -111  LockClr

                org     OVERLAY_START

OvcLXXXXXX      or      opc,#%0000_0100                 ' Set LockSet / LockClr instruction
                
                call    #PopLhs                         ' Get lock id

                hubop   lhs,opc                         ' lhs = 0..7, C := previous setting
                muxc    lhs,k_FFFF_FFFF                 ' True or false depending on previous setting
          
                and     opc,#%0000_0100 WZ,NR           ' Is it a function ?
        IF_Z    jmp     #Push                           ' Yes
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $3C     ?                                                                                       |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START

OvcUNKNOWN
jmp #NotImplemented

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $3D     REG     r[]                                                                             |
' |     $3E     REG     r[..]                                                                           |
' |     $3F     REG     r                                                                               |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; a := INA
'               ====                                    ; OUTA := a
'               ====                                    ; OUTA++
'               ====                                    ; a := CogId
'               
'               0018    3F 92                             PUSH     INA
'               001A    41                                POP      VAR+0.LONG
'
'               001B    40                                PUSH     VAR+0.LONG
'               001C    3F B4                             POP      OUTA
'
'               001E    3F D4 2E                          USING    OUTA POSTINC
'
'               0021    3F 89                   S3        PUSH     COGID
'               0023    41                                POP      VAR+0.LONG
'
'               ====                                    ; a := INA[ 0 ]
'               ====                                    ; OUTA[ 0 ] := a
'               ====                                    ; OUTA[ 0 ]++
'               
'               0018    35                                PUSH     0
'               0019    3D 92                             PUSH     INA[]
'               001B    41                                POP      VAR+0.LONG
'
'               001C    40                                PUSH     VAR+0.LONG
'               001D    35                                PUSH     0
'               001E    3D B4                             POP      OUTA[]
'
'               0020    35                                PUSH     0
'               0021    3D D4 28                          USING    OUTA[] POSTINC

'               ====                                    ; a := INA[ 4 .. 7 ]
'               ====                                    ; OUTA[ 4.. 7 ] := a
'               ====                                    ; OUTA[ 4.. 7]++
'               
'               0018    37 01                             PUSH     4
'               001A    37 22                             PUSH     7
'               001C    3E 92                             PUSH     INA[..]
'               001E    41                                POP      VAR+0.LONG
'
'               001F    40                                PUSH     VAR+0.LONG
'               0020    37 01                             PUSH     4
'               0022    37 22                             PUSH     7
'               0024    3E B4                             POP      OUTA[..]
'
'               0026    37 01                             PUSH     4
'               0028    37 22                             PUSH     7
'               002A    3E D4 28                          USING    OUTA[..] POSTINC

'                **
'               1001 rrrr       PUSH  r[..]
'               1011 rrrr       POP   r[..]
'               1100 rrrr       USING r[..]
                
                org     OVERLAY_START

OvcREG_LMM      cmp     opc,#$3F WZ, NR                 ' If REG r then full width ...

  IF_Z          mov     rhs,#0                          ' r[31..0]
  IF_Z          mov     lhs,#31
        
  IF_NZ         call    #PopRhs                         ' Get second or only bit number
  IF_NZ         shr     opc,#1 WC, NR
  IF_NZ_AND_NC  call    #PopLhs                         ' Get first bit number

                mov     bitRev,k_FFFF_FFFF              ' No bit reversing by default

                call    #SortLhsRhs                     ' Set lhs = lsb number, rhs = msb number

        IF_B    mov     bitRev,#31                      ' We swapped bit number direction so set bitRev
        IF_B    sub     bitRev,rhs                      ' ... bitRev := 31 - (rhs-lhs)
        IF_B    add     bitRev,lhs                      ' ... bitRev := 31 - rhs + lhs

                mov     bitMsk,k_FFFF_FFFF              ' Determine mask, example [7..4], lhs=4, rhs=7
                neg     rhs,rhs                         ' -7
                add     rhs,#31                         ' 31 + (-7) = 24
                add     rhs,lhs                         ' 24 + 4 = 28
                shr     bitMsk,rhs                      ' Gives $0000_000F
                shl     bitMsk,lhs                      ' Gives $0000_00F0

                mov     bitShf,lhs                      ' Set the bit shift to align lsb ( >> 4 )
               
                rdbyte  opc,pc                          ' Determine what we're actually doing
                add     pc,#1
                
                mov     opr,opc                         ' Set the operand - Special cases become at $00F
                and     opr,#$00F
                        
                cmp     opc,#$89 WZ, NR                 ' ... such as Cog Id
        IF_NZ   or      opr,#$1F0
        
                shr     opc,#5
                or      opc,#%0110_0000                 ' Force the Get/Put to use mov's not rdxxxx
                
                jmp     #Access_Spr_Here
         
DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Using Operator Overlays                                                                         *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |     $02     RPTINCJ repeat <var> from <expr> to <expr>                                              |
' |     $06     RPTADDJ repeat <var> from <expr> to <expr> step <expr>                                  |
' `-----------------------------------------------------------------------------------------------------'
'
'               ====                                    ; repeat a from 1 to 3
'               ====                                    ;   b := 1
'               
'               0018    36                                PUSH     1
'               0019    41                                POP      VAR+0.LONG
'               001A    36                      J4          PUSH   1
'               001B    45                                  POP    VAR+4.LONG
'               001C    36                                PUSH     1
'               001D    37 21                             PUSH     3
'               001F    42 02 78                          USING    VAR+0.LONG RPTINCJ -8

'               ====                                    ; repeat a from 1 to 3 step 2
'               ====                                    ;   b := 1
'               
'               0018    36                                PUSH     1
'               0019    41                                POP      VAR+0.LONG
'               001A    36                      J4          PUSH     1
'               001B    45                                  POP      VAR+4.LONG
'               001C    37 00                             PUSH     2
'               001E    36                                PUSH     1
'               001F    37 21                             PUSH     3
'               0021    42 06 76                          USING    VAR+0.LONG RPTADDJ -10

                org     OVERLAY_START

OvrRPTXXXJ      mov     acc,lhs                         ' Put index variable's value in acc

                call    #PopRhs                         ' Pop the range - lhs initial value
                call    #PopLhs                         '                 rhs final value

                and     opc,#%0000_0100 WZ, NR          ' Is it Inc or Add ?
        IF_Z    mov     adr,#1                          ' Inc
        IF_NZ   rdlong  adr,sp                          ' Add, Pop the step value
        IF_NZ   sub     sp,#4
                
                cmps    rhs,lhs WC, WZ, NR              ' Is rhs => lhs
        IF_AE   add     acc,adr                         ' Yes, increment
        IF_B    sub     acc,adr                         ' No,  decrement

                call    #GetAddress                     ' Get branch address
          
                call    #SortLhsRhs                     ' Sort lhs and rhs into ascending order

                cmps    acc,lhs WC, WZ, NR              ' Is index < minimum ?
        IF_B    jmp     #:Done                          ' Yes, finished

                cmps    acc,rhs WC, WZ, NR              ' Is index > maximum ?
        IF_A    jmp     #:Done                          ' Yes, finished

                add     pc,adr                          ' Repeat the loop
                
:Done           mov     lhs,acc                         ' Restore index variable's value

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $08     FWDRAND ? lhs                                                                           |
' `-----------------------------------------------------------------------------------------------------'

'               lhs := ( lhs <- 4 ) + 2531011

                org     OVERLAY_START
                
OvrFWDRAND      rol     lhs,#4
                add     lhs,:k_2531011
                jmp     #CommonReturn

:k_2531011      long    2531011

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $0C     REVRAND lhs ?                                                                           |
' `-----------------------------------------------------------------------------------------------------'

'               lhs := ( lhs - 2531011 ) -> 4

                org     OVERLAY_START
                
OvrREVRAND      sub     lhs,:k_2531011
                ror     lhs,#4
                jmp     #CommonReturn

:k_2531011      long    2531011

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $10     PRESEXB ~lhs                                                                            |
' |     $14     PRESEXW ~~lhs                                                                           |
' `-----------------------------------------------------------------------------------------------------'
'
'               ---1 0000 << 1 = ---0 0000 ^ ---1 1000 = ---1 1000 = 24     Byte shift
'               ---1 0100 << 1 = ---0 1000 ^ ---1 1000 = ---1 0000 = 16     Word shift

                org     OVERLAY_START
                
OvrPRESEXX      shl     opc,#1
                xor     opc,#%0001_1000
                shl     lhs,opc                 ' Byte << 24,  Word << 16
                sar     lhs,opc                 ' Byte ~> 24,  Word ~> 16

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

DAT

' *******************************************************************************************************
' *                                                                                                     *
' *     Unary Operator Overlays                                                                         *
' *                                                                                                     *
' *******************************************************************************************************

' .-----------------------------------------------------------------------------------------------------.
' |     $EF     BIT_REV lhs := lhs >< rhs                                                               |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvrBIT_REV      neg     rhs,rhs
                add     rhs,#32
                rev     lhs,rhs

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $F1     BIT_GET lhs := >| rhs                                                                   |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvrBIT_GET      mov     lhs,#0
:Loop           tjz     rhs,#CommonReturn
                shr     rhs,#1
                add     lhs,#1
                jmp     #:Loop

                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $F4     MPY     lhs := lhs * rhs                                                                |
' |     $F5     MPY_BIG lhs := lhs ** rhs                                                               |
' `-----------------------------------------------------------------------------------------------------'
'
'               lhsTop:lhsBot := || lhs
'                      rhsBot := || rhs
'               accTop:accBot := 0
'
'               while rhs
'                 if rhs & 1
'                   accTop:accBot += lhsTop:lhsBot
'                   lhsTop:lhsBot := lhsTop:lhsBot << 1
'                 rhs := rhs >> 1

                org     OVERLAY_START
                
OvrMPY          abs     lhs,lhs WC                      ' Sets C if was negative
                muxc    :signBit,#1

                abs     rhs,rhs WC                      ' Sets C if was negative
        IF_C    xor     :signBit,#1

                mov     :lhsTop,#0
                mov     :accBot,#0
                mov     :accTop,#0
                
:Loop           shr     rhs,#1 WC, WZ                   ' if rhs & 1

        IF_C    add     :accTop,:lhsTop                 '   accTop:accBot += lhsTop:lhsBot
        IF_C    add     :accBot,:lhsBot WC
        IF_C    add     :accTop,#1

                shl     :lhsBot,#1 WC                   '   lhsTop:lhsBot := lhsTop:lhsBot << 1
                rcl     :lhsTop,#1
       
        IF_NZ   jmp     #:Loop

' This is the old version ...

'               If signbit is set we need to negate accTop:acc. We do that by the standard trick of
'               inverting and adding one. There may be a more efficient way of doing this; using neg
'               etc.

'                shr     :signBit,#1 WC, NR
'        IF_C    xor     :accTop,k_FFFF_FFFF
'        IF_C    xor     :accBot,k_FFFF_FFFF
'        IF_C    add     :accBot,#1 WC
'        IF_C    add     :accTop,#1

' Not tested ...

                shr     :signBit,#1 WC, NR
    IF_C        neg     :accBot,:accBot WZ
    IF_C        neg     :accTop,:accTop
    IF_C_AND_NZ sub     :accTop,#1
                
                shr     opc,#1 WC, NR
        IF_NC   mov     lhs,:accBot                     ' MPY
        IF_C    mov     lhs,:accTop                     ' MPY_BIG

:org            org     REGISTERS
:lhsTop         res     1
:accTop         res     1
:signBit        res     1
                fit     LAST_REGISTER+1

                org     lhs
:lhsBot         res     0
                org     acc
:accBot         res     0

                org     :org
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

' .-----------------------------------------------------------------------------------------------------.
' |     $F6     DIV     lhs := lhs / rhs                                                                |
' |     $F7     MOD     lhs := lhs // rhs                                                               |
' `-----------------------------------------------------------------------------------------------------'

                org     OVERLAY_START
                
OvrDIV          abs     rhs,rhs WC, WZ                  ' Sets C if was negative
        IF_Z    mov     lhs,#0
        IF_Z    jmp     #CommonReturn
                muxc    :signBit,#1

                abs     lhs,lhs WC                      ' Sets C if was negative
        IF_C    xor     :signBit,#1

                mov     acc,#0
                
                mov     :count,#1

:ShlLoop        and     rhs,:k_4000_0000 WZ,NR
        IF_Z    shl     rhs,#1
        IF_Z    add     :count,#1
        IF_Z    jmp     #:ShlLoop

:SubLoop        shl     acc,#1
                cmps    lhs,rhs WC, WZ, NR
        IF_AE   sub     lhs,rhs
        IF_AE   or      acc,#1
                shr     rhs,#1
                djnz    :count,#:SubLoop

'               acc   = result
'               lhs   = remainder

                shr     opc,#1 WC, NR
        IF_NC   mov     lhs,acc                         ' DIV

'               If sign bit was set we need to negate back

                shr     :signBit,#1 WC, NR
        IF_C    neg     lhs,lhs

                jmp     #CommonReturn

:k_4000_0000    long    $4000_0000

:org            org     REGISTERS
:count          res     1
:signBit        res     1
                fit     LAST_REGISTER+1
                org     :org
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END
                            
' .-----------------------------------------------------------------------------------------------------.
' |     $F8     SQRT    lhs := ^^ rhs                                                                   |
' `-----------------------------------------------------------------------------------------------------'
'
'               lhs := || lhs
'               root := 0
'               if lhs > $1_0000
'                 bit := |< 30
'               else
'                 bit := |< 14
'               repeat
'                 trial := root + bit
'                 if lhs => trial
'                   lhs -= trial
'                   root := trial + bit
'                 root >>= 1
'                 bit >>= 2 
'               while bit
'               result := root
  
                org     OVERLAY_START
                
OvrSQRT         abs     lhs,lhs                         ' lhs := || lhs

                mov     :root,#0                        ' root := 0
                
                mov     :bit,#1
                andn    lhs,k_0000_FFFF WZ, NR          ' if lhs > $1_0000
        IF_NZ   shl     :bit,#30                        '   bit := |< 30, else
        IF_Z    shl     :bit,#14                        '   bit := |< 14
                                                        ' repeat
:Loop           mov     :trial,:root                    '   trial := root + bit
                add     :trial,:bit

                cmps    lhs,:trial WZ, WC, NR           '   if lhs => trial
        IF_AE   sub     lhs,:trial                      '     lhs -= trial
        IF_AE   mov     :root,:trial                    '     root := trail + bit
        IF_AE   add     :root,:bit     

                shr     :root,#1                        '   root >>= 1
                shr     :bit,#2                         '   bit >>= 2

                tjnz    :bit,#:Loop                     '  while bit

                mov     lhs,:root                       ' result := root

:org            org     REGISTERS
:root           res     1
:trial          res     1
:bit            res     1
                fit     LAST_REGISTER+1
                org     :org
                
                long    0[ 0 #> (OVERLAY_END-$) <# 1 ]
                fit     OVERLAY_END

EndOfCog        res     0

' *******************************************************************************************************
' *                                                                                                     *
' *     End of Spin VM Interpreter                                                                      *
' *                                                                                                     *
' *******************************************************************************************************

CON


CON

{
PRI StackTest | i

  cognew( Level1, @stack )
  
  PauseMs(1000)

  repeat i from 0 to 12
    TxByte($0D)
    TxHex( @stack[i],4)
    TxByte(" ")
    TxByte(":")
    TxByte(" ")
    TxHex(stack[i]>> 16,4)
    TxByte(" ")
    TxHex(stack[i],4)
  repeat

PRI Level1
  Level2($111111,$222222)

PRI Level2(v1,v2)
  Level3($333333,$444444)

PRI Level3(v1,v2)
  Level4($555555,$666666)

PRI Level4(v1,v2)
  waitpeq(1,$FFFF_FFFF,0)
}

{
PRI TimeTest

  repeat
    TxByte( $01 )
    TxByte( $0D )
    TxDec( long[$7000] )
    if long[$7000] => 10_000_000
      TxText(STRING("  FINISHED"))

PRI Tick
  repeat
    TickSub
   
PRI TickSub
  long[$7000]++
}

